"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


from qrisp.core import QuantumVariable
import numpy as np
import sympy as sp
from qrisp.misc import gate_wrap
from qrisp.arithmetic import polynomial_encoder, lt, gt, leq, geq, eq, neq, inpl_mult, sbp_add

def signed_int_iso(x, n):
    if int(x) < -2**n or int(x) >= 2**n:
        raise Exception("Applying signed integer isomorphism resulted in overflow")
    
    if x >= 0:
        return x%2**n
    else:
        return -abs(x)%2**(n+1)

def signed_int_iso_inv(y, n):
    y = y%2**(n+1)
    if y < 2**n:
        return y
    else:
        return -2**(n+1) + y


#Truncates a polynomial of the form p(x) = 2**k_0*x*i_0 + 2**k_1*x**i_1 ...
#where every summand where the power of the coefficients does not lie in the interval
#trunc bounds is removed
def trunc_poly(poly, trunc_bounds):
    
    #Convert to sympy polynomial
    poly = sp.poly(poly)

    #Clip upper bound
    poly = poly.trunc(2.**(trunc_bounds[1]))
    
    #Clip lower bound
    poly = poly/2.**trunc_bounds[0]
    poly = poly - sp.poly(poly).trunc(1)
    poly = poly*2.**trunc_bounds[0]
    
    return poly.expr.expand()    



class QuantumFloat(QuantumVariable):
    r"""
    This subclass of :ref:`QuantumVariable` can represent floating point numbers (signed and unsigned) up to an arbitrary precision.
    
    The technical details of the employed arithmetic can be found in this `article <https://ieeexplore.ieee.org/document/9815035>`_. 
    
    To create a QuantumFloat we call the constructor:
    
    >>> from qrisp import QuantumFloat
    >>> qf_0 = QuantumFloat(3, -1, signed = False)
    
    Here, the 3 indicates the amount of mantissa qubits and the -1 indicates the exponent.
    
    For unsigned QuantumFloats, the decoder function is given by
    
    .. math::
        
        f_{k}(i) = i2^{k}
    
    Where $k$ is the exponent.
    
    We can check which values can be represented:
    
    >>> for i in range(2**qf_0.size): print(qf_0.decoder(i))
    0.0
    0.5
    1.0
    1.5
    2.0
    2.5
    3.0
    3.5
    
    We see $2^3 = 8$ values, because we have 3 mantissa qubits. The exponent is -1, implying the precision is $0.5 = 2^{-1}$.
    
    For signed QuantumFloats, the decoder function is
    
    .. math::
        
        f_{k}^{n}(i) = \begin{cases} i2^{k} & \text{if } i < 2^n \\ (i - 2^{n+1})2^k & \text{else} \end{cases}
    
    Where $k$ is again, the exponent and $n$ is the mantissa size.
    
    
    Another example:
    
    >>> qf_1 = QuantumFloat(2, -2, signed = True)
    >>> for i in range(2**qf_1.size): print(qf_1.decoder(i))
    0.0
    0.25
    0.5
    0.75
    -1.0
    -0.75
    -0.5
    -0.25
    
    Here, we have $2^2 = 4$ values and their signed equivalents. Their precision is $0.25 = 2^{-2}$.
    
    
    **Arithmetic**
    
    Many operations known from classical arithmetic work for QuantumFloats in infix notation.
    
    Addition:
    
    >>> qf_0[:] = 1.5
    >>> qf_1[:] = 0.25
    >>> qf_3 = qf_0 + qf_1
    >>> print(qf_3)
    {1.75: 1.0}
    
    Subtraction:
    
    >>> qf_4 = qf_0 - qf_3
    >>> print(qf_4)
    {-0.25: 1.0}
    
    Multiplication:
        
    >>> qf_5 = qf_4 * qf_1
    >>> print(qf_5)
    {-0.0625: 1.0}
    
    And even division:
        
    >>> qf_6 = QuantumFloat(3)
    >>> qf_7 = QuantumFloat(3)
    >>> qf_6[:] = 7
    >>> qf_7[:] = 2
    >>> qf_8 = qf_6/qf_7
    >>> print(qf_8)
    {3.5: 1.0}
    
    Floor division:
        
    >>> qf_9 = qf_6//qf_7
    >>> print(qf_9)
    {3.0: 1.0}
    
    Inversion:
    
    >>> qf_10 = QuantumFloat(3, -1)
    >>> qf_10[:] = 3.5
    >>> qf_11 = qf_10**-1
    >>> print(qf_11)
    {0.25: 1.0}
    
    Note that the latter is only an approximate result. This is because in many cases, the results of division can not be stored in a finite amount of qubits, forcing us to approximate.
    To get a better approximation we can use the :meth:`q_div <qrisp.q_div>` and :meth:`qf_inversion <qrisp.qf_inversion>` functions and specify the precision:
    
    >>> from qrisp import q_div, qf_inversion
    >>> qf_12 = QuantumFloat(3)
    >>> qf_12[:] = 1
    >>> qf_13 = QuantumFloat(3)
    >>> qf_13[:] = 7
    >>> qf_14 = q_div(qf_12, qf_13, prec = 6)
    >>> print(qf_14)
    {0.140625: 1.0}
    
    Comparing with the classical result (0.1428571428):
    
    >>> 1/7 - 0.140625
    0.002232142857142849
    
    We see that the result is inside the expected precision of $2^{-6} =  0.015625$.
    
    
    **In-place Operations**
    
    Further supported operations are inplace addition, subtraction (with both classical and quantum values):
        
    >>> qf_15 = QuantumFloat(4, signed = True)
    >>> qf_15[:] = 4
    >>> qf_16 = QuantumFloat(4)
    >>> qf_16[:] = 3
    >>> qf_15 += qf_16
    >>> print(qf_15)
    {7: 1.0}
    >>> qf_15 -= 2
    >>> print(qf_15)
    {5: 1.0}
    
    .. warning::
        Additions that would result in overflow, raise no errors. Instead the additions are performed `modular <https://en.wikipedia.org/wiki/Modular_arithmetic>`_.
        
        >>> a = QuantumFloat(3)
        >>> a += 9
        >>> print(a)
        {1: 1.0}
    
    For inplace multiplications, only classical integers are allowed:
    
    >>> qf_15 *= -3
    >>> print(qf_15)
    {-15: 1.0}
    
    .. note::
        In-place multiplications can change the mantissa size to prevent overflow errors. If you want to prevent this behavior, look into :meth:`inpl_mult <qrisp.inpl_mult>`.
        
        >>> qf.qf_15.size
        7
    
    **Bitshifts**
    
    Bitshifts can be executed for free (ie. not requiring any quantum gates). We can either use the :meth:`exp_shift <qrisp.QuantumFloat.exp_shift>` method or use the infix operators. Note that the bitshifts work in-place.
    
    
    >>> qf_15.exp_shift(3)
    >>> print(qf_15)
    {40: 1.0}
    >>> qf_15 >> 5
    >>> print(qf_15)
    {1.25: 1.0}
    
    **Comparisons**
    
    QuantumFloats can be compared to Python floats using the established operators. The return values are :ref:`QuantumBools <QuantumBool>`:
        
    >>> from qrisp import h
    >>> qf_16 = QuantumFloat(4)
    >>> h(qf_16[2])
    >>> print(qf_16)
    {0: 0.5, 4: 0.5}
    >>> comparison_qbl_0 = (qf_16 < 4 )
    >>> print(comparison_qbl_0)
    {False: 0.5, True: 0.5}

    Comparison to other QuantumFloats also works:
        
    >>> qf_17 = QuantumFloat(3)
    >>> qf_17[:] = 4
    >>> comparison_qbl_1 = (qf_17 == qf_16)
    >>> comparison_qbl_1.qs.statevector()
    sqrt(2)*(|0>*|True>*|4>*|False> + |4>*|False>*|4>*|True>)/2
    
    The first tensor factor containing a boolean value is corresponding to ``comparison_qbl_0`` and the second one is ``comparison_qbl_1``.
    
    """
    
    
    def __init__(self, msize, exponent = 0, qs = None, name = None, signed = False):
        
        #Boolean to indicate if the float is signed
        self.signed = signed
        
        #Exponent
        self.exponent = exponent
        
        #Size of the mantissa
        self.msize = msize
        
        #Array that consists of (log2(min), log2(max)) where min and max are the minimal and maximal values
        #of the absolutes that the QuantumFloat can represent
        self.mshape = np.array([exponent, exponent + msize])
        
        #Initialize QuantumVariable
        if signed:
            super().__init__(msize + 1, qs, name = name)
        else:
            super().__init__(msize, qs, name = name)
            
        
    
        
        
    #Define outcome_labels
    def decoder(self, i):
        if self.signed:
            
            res = signed_int_iso_inv(i, self.size-1)*2.**self.exponent
        else:
            res =  i*2**self.exponent
            
        if self.exponent >= 0:
            return int(res)
        else:
            return res
        
    # def encoder(self, i):
    #     if self.signed:
    #         res = int(signed_int_iso(i/2**self.exponent, self.size-1))
    #     else:
    #         res =  int(i/2**self.exponent)
            
    #     return res
        
        
    
    def sb_poly(self, m = 0):
        """
        Returns the semi-boolean polynomial of this `QuantumFloat` where `m` specifies the image extension parameter.
        
        For the technical details we refer to: https://ieeexplore.ieee.org/document/9815035
        

        Parameters
        ----------
        m : int, optional
            Image extension parameter. The default is 0.

        Returns
        -------
        Sympy expression
            The semi-boolean polynomial of this QuantumFloat.
            
        Examples
        --------
        
        >>> from qrisp import QuantumFloat
        >>> x = QuantumFloat(3, -1, signed = True, name = "x")
        >>> print(x.sb_poly(5))
        0.5*x_0 + 1.0*x_1 + 2.0*x_2 + 28.0*x_3
        
        """
        
        if m == 0:
            m = self.size
        
        symbols = [sp.symbols(self.name + "_" + str(i)) for i in range(self.size)]
        
        poly = sum([2.**(i)*symbols[i] for i in range(self.size)])
        
        
        if self.signed:
                poly += (2.**(m+1) - 2.**(self.size))*symbols[-1]
        
        return 2**self.exponent*poly
    
    
    def encode(self, encoding_number, rounding = False):
        
        
        if rounding:
            #Round value to closest fitting number    
            outcome_labels = [self.decoder(i) for i in range(2**self.size)]
            encoding_number = outcome_labels[np.argmin(np.abs(encoding_number-np.array(outcome_labels)))]
    
        super().encode(encoding_number)
        
    @gate_wrap(permeability = "args", is_qfree = True)
    def __mul__(self, other):
        
        if isinstance(other, QuantumFloat):
                return q_mult(self, other)
        elif isinstance(other, int):
            
            bit_shift = 0
            while not other%2:
                other = other>>1
                bit_shift += 1                
            
            if self.signed or other < 0:
                output_qf = QuantumFloat(self.msize + int(np.ceil(np.log2(abs(other)))),
                                         self.exponent, 
                                         signed = True)
            else:
                output_qf = QuantumFloat(self.msize + int(np.ceil(np.log2(abs(other)))), 
                                         self.exponent, 
                                         signed = False)
            
            polynomial_encoder([self], output_qf, other*sp.Symbol("x"))
            
            output_qf<<bit_shift
            
            return output_qf
        else:
            raise Exception("QuantumFloat multiplication for type " + str(type(other)) + ""
                            " not implemented (available are QuantumFloat and int)")
    
    @gate_wrap(permeability = "args", is_qfree = True)
    def __add__(self, other):
        if isinstance(other, QuantumFloat):
            return sbp_add(self, other)
        elif isinstance(other, (int, float)):
            res = self.duplicate(init = True)
            res += other
            return res            
        else:
            raise Exception("Addition with type " + str(type(other)) + " not implemented")
    
    @gate_wrap(permeability = "args", is_qfree = True)
    def __sub__(self, other):
        from qrisp.arithmetic import sbp_sub
        if isinstance(other, QuantumFloat):
            
            return sbp_sub(self, other)
        elif isinstance(other, (int, float)):
            
            res = self.duplicate(init = True)
            res -= other
            return res
        else:
            raise Exception("Subtraction with type " + str(type(other)) + " not implemented")
            
    __radd__ = __add__
    __rmul__ = __mul__
    
    @gate_wrap(permeability = "args", is_qfree = True)
    def __rsub__(self, other):
        from qrisp.arithmetic import sbp_sub
        from qrisp import x
        if isinstance(other, QuantumFloat):
            return sbp_sub(other, self)
        elif isinstance(other, (int, float)):
            
            res = self.duplicate(init = True)
            if not res.signed:
                res.add_sign()
            x(res)
            res -= other - 2**res.exponent
            return res
        else:
            raise Exception("Subtraction with type " + str(type(other)) + " not implemented")
        
    @gate_wrap(permeability = "args", is_qfree = True)
    def __truediv__(self, other):
        from qrisp.arithmetic import q_div
        
        return q_div(self, other)
    
    @gate_wrap(permeability = "args", is_qfree = True)
    def __floordiv__(self, other):
        
        if self.signed or other.signed:
            raise Exception("Floor division not implemented for signed QuantumFloats")
        
        if self.exponent < 0 or other.exponent < 0:
            raise Exception("Tried to perform floor division on non-integer QuantumFloats")
        from qrisp.arithmetic import q_div
        
        return q_div(self, other, prec = 0)
    
    @gate_wrap(permeability = "args", is_qfree = True)
    def __pow__(self, power):
        
        if power != -1:
            raise Exception("Currently the only supported power is -1")
        
        from qrisp.arithmetic import qf_inversion
        
        return qf_inversion(self)
    
    @gate_wrap(permeability = [1], is_qfree = True)
    def __iadd__(self, other):
        
        if isinstance(other, QuantumFloat):
            input_qf_list = [other]
            poly = sp.symbols("x")
            
            polynomial_encoder(input_qf_list, self, poly)
            
        
        elif isinstance(other, (int, float)):
            # self.incr(other)
            
            if not int(other/2**self.exponent) == other/2**self.exponent:
                raise Exception("Tried to perform in-place addition with invalid number. QuantumFloat precision too low.")
            
            input_qf_list = []
            poly = sp.sympify(other)
            
            polynomial_encoder(input_qf_list, self, poly)

        else:
            raise Exception("In-place addition for type " + str(type(other)) + " not implemented")
        
        return self
    
    @gate_wrap(permeability = [1], is_qfree = True)
    def __isub__(self, other):
        
        if isinstance(other, QuantumFloat):
            input_qf_list = [other]
            poly = -sp.symbols("x")
        
            polynomial_encoder(input_qf_list, self, poly)
            
        elif isinstance(other, (int, float)):
            
            if not int(other/2**self.exponent) == other/2**self.exponent:
                raise Exception("Tried to perform in-place subtraction with invalid number. QuantumFloat precision too low.")
            
            input_qf_list = []
            poly = -sp.sympify(other)
        
            polynomial_encoder(input_qf_list, self, poly)
            
        else:
            raise Exception("In-place substraction for type " + str(type(other)) + " not implemented")
        
        return self
            
    
    @gate_wrap(permeability=[], is_qfree = True)
    def __imul__(self, other):
        
        inpl_mult(self, other)
        
        return self
        
    def __rshift__(self, k):
        self.exp_shift(-k)
        return self
    
    def __lshift__(self, k):
        self.exp_shift(k)
        return self
    
    def __lt__(self, other):
        
        if not isinstance(other, (QuantumFloat, int, float)):
            raise Exception(f"Comparison with type {type(other)} not implemented")
        
        return lt(self, other)
        
    def __gt__(self, other):
        
        if not isinstance(other, (QuantumFloat, int, float)):
            raise Exception(f"Comparison with type {type(other)} not implemented")
        
        return gt(self, other)
        

    def __le__(self, other):
        
        if not isinstance(other, (QuantumFloat, int, float)):
            raise Exception(f"Comparison with type {type(other)} not implemented")
        
        return leq(self, other)

    def __ge__(self, other):
        
        if not isinstance(other, (QuantumFloat, int, float)):
            raise Exception(f"Comparison with type {type(other)} not implemented")
        
        return geq(self, other)
        
    def __eq__(self, other):
        
        if not isinstance(other, (QuantumFloat, int, float)):
            raise Exception(f"Comparison with type {type(other)} not implemented")
        
        return eq(self, other)

    def __ne__(self, other):
        
        if not isinstance(other, (QuantumFloat, int, float)):
            raise Exception(f"Comparison with type {type(other)} not implemented")
        
        return neq(self, other)
        
    def exp_shift(self, shift):
        
        if not isinstance(shift, int):
            raise Exception("Tried to shift QuantumFloat exponent by non-integer value")
        
        self.exponent += shift
        self.mshape = self.mshape + shift
        
    def reduce(self, qubits, verify = False):
        QuantumVariable.reduce(self, qubits, verify)
        
        
        try:
            self.mshape[1] -= len(qubits)
            self.msize -= len(qubits)
        except TypeError:
            self.mshape[1] -= 1
            self.msize -= 1
            
        
            
    def extend(self, amount, position = -1):
        QuantumVariable.extend(self, amount, position = position)
        
        self.mshape[1] += amount
        self.msize += amount
    
    def add_sign(self):
        """
        Turns an unsigned QuantumFloat into its signed version.

        Raises
        ------
        Exception
            Tried to add sign to signed QuantumFloat.

        Examples
        --------
        
        >>> from qrisp import QuantumFloat
        >>> qf = QuantumFloat(4)
        >>> qf.signed
        False
        >>> qf.add_sign()
        >>> qf.signed
        True

        """
        
        if self.signed:
            raise Exception("Tried to add sign to signed QuantumFloat")
        
        self.extend(1, self.size-1)
        self.mshape[1] -= 1
        self.msize -= 1
        self.signed = True
        
    def sign(self):
        """
        Returns the sign qubit. 
        
        This qubit is in state $\ket{1}$ if the QuantumFloat holds a negative value and in state $\ket{0}$ otherwise.
        
        For more information about the encoding of negative numbers check our `paper <https://ieeexplore.ieee.org/document/9815035>`_.
        
        .. warning::
            
            Performing an X gate on this qubit does not flip the sign! Use inplace multiplication instead
            
            >>> from qrisp import QuantumFloat
            >>> qf = QuantumFloat(3, signed = True)
            >>> qf[:] = 3
            >>> qf *= -1
            >>> print(qf)
            {-3: 1.0}

        Raises
        ------
        Exception
            Tried to retrieve sign qubit of unsigned QuantumFloat.

        Returns
        -------
        Qubit
            The qubit holding the sign.
            
        Examples
        --------
        
        We create a QuantumFloat, initiate a state that has probability 2/3 of being negative and entangle a QuantumBool with the sign qubit.
        
        >>> from qrisp import QuantumFloat, QuantumBool, cx
        >>> qf = QuantumFloat(4, signed = True)
        >>> n_amp = 1/3**0.5
        >>> qf[:] = {-1 : n_amp, -2 : n_amp, 1 : n_amp}
        >>> qbl = QuantumBool()
        >>> cx(qf.sign(), qbl)
        >>> print(qbl)
        {True: 0.6667, False: 0.3333}

        """
        if not self.signed:
            raise Exception("Tried to retrieve sign qubit of unsigned QuantumFloat")
            
        return self[-1]
    
    
    def init_from(self, other, ignore_rounding_errors = False, ignore_overflow_errors = False):
        copy_qf(self, other, ignore_rounding_errors = ignore_rounding_errors, ignore_overflow_errors= ignore_overflow_errors)
    
    def incr(self, x = None):
        from qrisp.arithmetic.incrementation import increment
        if x == None:
            x = 2**self.exponent
        increment(self, x)

    def __hash__(self):
        return id(self)
    
    def significant(self, k):
        """
        Returns the qubit with significance $k$.

        Parameters
        ----------
        k : int
            The significance.

        Raises
        ------
        Exception
            Tried to retrieve invalid significant from QuantumFloat

        Returns
        -------
        Qubit
            The Qubit with significance $k$.
            
        Examples
        --------
        
        We create a QuantumFloat and flip a qubit of specified significance.
        
        >>> from qrisp import QuantumFloat, x
        >>> qf = QuantumFloat(6, -3)
        >>> x(qf.significant(-2))
        >>> print(qf)
        {0.25: 1.0}
        
        The qubit with significance $-2$ corresponds to the value $0.25 = 2^{-2}$.
        
        >>> x(qf.significant(2))
        {4.25: 1.0}
        
        The qubit with significance $2$ corresponds to the value $4 = 2^{2}$.

        """
        
        sig_list = list(range(self.mshape[0], self.mshape[1]))
        
        if not k in sig_list:
            raise Exception(f"Tried to retrieve invalid significant {k} from QuantumFloat with mantissa shape {self.mshape}")
        
        return self[sig_list.index(k)]
        
   
    def truncate(self, x):
        """
        Receives a regular float and returns the float that is closest to the input but can still be encoded.

        Parameters
        ----------
        x : float
            A float that is supposed to be truncated.

        Returns
        -------
        float
            The truncated float.
            
        Examples
        --------
        
        We create a QuantumFloat and round a value to fit the encoder and subsequently initiate:
        
        >>> from qrisp import QuantumFloat
        >>> qf = QuantumFloat(4, -1)
        >>> value = 0.5102341
        >>> qf[:] = value
        Exception: Value 0.5102341 not supported by encoder.
        >>> rounded_value = qf.truncate(value)
        >>> rounded_value
        0.5
        >>> qf[:] = rounded_value
        >>> print(qf)
        {0.5: 1.0}

        """
        decoder_values = np.array([self.decoder(i) for i in range(2**self.size)])
        
        return decoder_values[np.argmin(np.abs(decoder_values - x))]
        


from qrisp.misc import gate_wrap

def q_mult(factor_1, factor_2, target = None, method = "auto"):
    
    if method == "auto":
        
        if factor_1.reg == factor_2.reg:
            return q_mult(factor_1, factor_2, target, method = "sbp")
        else:
            return q_mult(factor_1, factor_2, target, method = "hybrid")

    elif method == "sbp":
        
        from qrisp import sbp_mult, merge
        if target is None:
            target = create_output_qf([factor_1, factor_2], op = "mul")
        
        merge([target, factor_1, factor_2])
        
        
        sbp_mult(factor_1, factor_2, target)
        return target
    
    elif method == "hybrid":
        from qrisp.arithmetic import hybrid_mult
        return hybrid_mult(factor_1, factor_2)

def create_output_qf(operands, op):
    
    if isinstance(op, sp.core.expr.Expr):
        from qrisp.arithmetic.poly_tools import expr_to_list
        
        expr_list = expr_to_list(op)
        
        
        for i in range(len(expr_list)):
            if not isinstance(expr_list[i][0], sp.Symbol):
                expr_list[i].pop(0)
                
        
        operands.sort(key = lambda x : x.name)
        
        
        def prod(iter):
            iter = list(iter)
            a = iter[0]
            for i in range(1, len(iter)):
                a *= iter[i]
                
            return a
        
        from sympy import Symbol, Abs, Poly
        
        poly = Poly(op)
        monom_list = [a*prod(x**k for x, k in zip(poly.gens, mon)) for a, mon in zip(poly.coeffs(), poly.monoms())]
        
        max_value_dic = {Symbol(qf.name) : 2.**qf.mshape[1] for qf in operands}
        min_value_dic = {Symbol(qf.name) : 2.**qf.mshape[0] for qf in operands}
                
        
        
        abs_poly = sum([Abs(monom) for monom in monom_list], 0)
        
        min_poly_value = min([float(Abs(monom).subs(min_value_dic)) for monom in monom_list])
        
        max_poly_value = float(abs_poly.subs(max_value_dic))
        
        
        min_sig = int(np.floor(np.log2(min_poly_value)))
        max_sig = int(np.ceil(np.log2(max_poly_value)))
        
        
        msize  = max_sig - min_sig
        exponent = min_sig
        
        
        signed = bool(sum([int(operand.signed) for operand in operands]))
        
        return QuantumFloat(msize, exponent = exponent, signed = signed)
    

    
    if op == "add":
        signed = operands[0].signed or operands[1].signed
        exponent = min(operands[0].exponent, operands[1].exponent)
        

        max_sig = int(np.ceil(np.log2(int(2**operands[0].mshape[1] + 2**operands[1].mshape[1]))))
        msize = max_sig - exponent + 1
        return QuantumFloat(msize, exponent, operands[0].qs, signed = signed, name = "add_res*")
    
    if op == "mul":
        signed = operands[0].signed or operands[1].signed
        
        if operands[0].reg == operands[1].reg and (operands[0].signed and operands[1].signed):
            signed = False
        
        return QuantumFloat(operands[0].msize + operands[1].msize + operands[0].signed*operands[1].signed, operands[0].exponent + operands[1].exponent, operands[0].qs, signed = signed, name = "mul_res*")
    
    if op == "sub":
        exponent = min(operands[0].exponent, operands[1].exponent)
        max_sig = int(np.ceil(np.log2(int(2**operands[0].mshape[1] + 2**operands[1].mshape[1]))))
        msize = max_sig - exponent + 1
        
        return QuantumFloat(msize, exponent, operands[0].qs, signed = True, name = "sub_res*")
    


#Initiates the value of qf2 into qf1 where qf1 has to hold the value 0
def copy_qf(qf1, qf2, ignore_overflow_errors = False, ignore_rounding_errors = False):
    
    #Lists that translate Qubit index => Significance
    qf1_sign_list = [qf1.exponent + i for i in range(qf1.size)]
    qf2_sign_list = [qf2.exponent + i for i in range(qf2.size)]
    
    
    #Check overflow/underflow
    if max(qf1_sign_list) < max(qf2_sign_list) and not ignore_overflow_errors:
        raise Exception("Copy operation would result in overflow (use ignore_overflow_errors = True)")
        
    if min(qf1_sign_list) > min(qf2_sign_list) and not ignore_rounding_errors:
        raise Exception("Copy operation would result in rounding (use ignore_rounding_errors = True)")
    
    qs = qf1.qs
    
    if qf2.signed:
        if not qf1.signed:
            raise Exception("Tried to copy signed into unsigend float")
        
        #Remove last entry from significance list (last qubit is the sign qubit)
        qf2_sign_list.pop(-1)
        qf1_sign_list.pop(-1)
        
        
    
    for i in range(len(qf1_sign_list)):
        
        #If we are in a realm where both floats have overlapping significance => CNOT into each other
        if qf1_sign_list[i] in qf2_sign_list:
            qf2_index = qf2_sign_list.index(qf1_sign_list[i])
            qs.cx(qf2[qf2_index], qf1[i])
            continue
        
        #Otherwise copy the sign bit into the bits of higher significance than qf2
        if qf1_sign_list[i] > max(qf2_sign_list) and qf2.signed:
            qs.cx(qf2[-1], qf1[i])
    
    #Copy the sign bit
    if qf2.signed:
        qs.cx(qf2[-1], qf1[-1])




# @adaptive_condition
# def less_than(qf_0, qf_1):
#     lt_qbl = QuantumBool()
#     if qf_0.signed:
#         qf_0 -= qf_1
#         cx(qf_1[-1], lt_qbl)
#         # The last qubit holds the sign of the QuantumFloat
#         qf_0 += qf_1
#         return lt_qbl
#     else:
#         temp_qf = qf_0 - qf_1
#         cx(temp_qf[-1], lt_qbl)
#         return lt_qbl

# @adaptive_condition
# def greater_than(qf_0, qf_1):
#     gt_qbl = QuantumBool()
#     if qf_0.signed:
#         qf_1 -= qf_0
#         # The last qubit holds the sign of the QuantumFloat
#         cx(qf_1[-1], gt_qbl)
#         qf_1 += qf_0
#         return gt_qbl
#     else:
#         temp_qf = qf_1 - qf_0
#         cx(temp_qf[-1], gt_qbl)
#         return gt_qbl

# @adaptive_condition
# def less_than_or_equal(qf_0, qf_1):
#     leq_qbl = QuantumBool()
#     if qf_0.signed:
#         qf_1 -= qf_0
#         # The last qubit holds the sign of the QuantumFloat
#         cx(qf_0[-1], leq_qbl)
#         qf_1 += qf_0
#         leq_qbl.flip()
#         return leq_qbl
#     else:
#         temp_qf = qf_1 - qf_0
        
#         cx(temp_qf[-1], leq_qbl)
#         leq_qbl.flip()
#         return leq_qbl

# @adaptive_condition
# def greater_than_or_equal(qf_0, qf_1):
#     geq_qbl = QuantumBool()
#     if hasattr(qf_0, 'signed') and qf_0.signed:
#         qf_0 -= qf_1
#         # The last qubit holds the sign of the QuantumFloat
#         cx(qf_0[-1], geq_qbl)
#         qf_0 += qf_1
#         geq_qbl.flip()
#         return geq_qbl
#     else:
#         temp_qf = qf_0 - qf_1
#         cx(temp_qf[-1], geq_qbl)
#         geq_qbl.flip()
#         return geq_qbl


# #TO-DO implement without arithmetic?
# @adaptive_condition
# def equal(qf_0, qf_1):
    
#     eq_qbl = QuantumBool()
    
#     if isinstance(qf_1, QuantumFloat):
#         if not qf_0.signed and not qf_1.signed:
#             if qf_0.exponent > qf_1.exponent:
#                 qf_0, qf_1 = qf_1, qf_0
                
#             min_sig = qf_0.mshape[0]
#             max_sig = max(qf_0.mshape[1], qf_1.mshape[1])
            
#             temp_qf = QuantumFloat(max_sig - min_sig)
            
#             cx(qf_0, temp_qf[:qf_0.size])
#             cx(qf_1, temp_qf[-qf_1.size:])
            
#             mcx(temp_qf, eq_qbl, ctrl_state = 0)
            
#             return eq_qbl
        
#     temp_qf = qf_0 - qf_1
        
#     mcx(temp_qf, eq_qbl, ctrl_state = 0)
    
#     return eq_qbl

# @adaptive_condition
# def always_true():
#     qbl = QuantumBool()
#     qbl.flip()
#     return qbl

# @adaptive_condition
# def always_false():
#     from qrisp import QuantumBool
#     qbl = QuantumBool()
#     return qbl






"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


from qrisp.qtypes.quantum_char import QuantumChar
from qrisp.core.quantum_array import QuantumArray


import numpy as np

nisq_init_quantum_char = QuantumChar(name = "nisq_q_char")
init_quantum_char = QuantumChar(nisq_char = False, name = "q_char")

class QuantumString(QuantumArray):
    """
    The QuantumString is the quantum equivalent of a string. It is implemented as a :ref:`QuantumArray` of :ref:`QuantumChars <QuantumChar>`.
    
    >>> from qrisp import QuantumString
    >>> q_str = QuantumString(size = len("hello world"))
    >>> q_str[:] = "hello world"
    >>> print(q_str)    
    {'hello world': 1.0}
    
    It is also possible to have a QuantumString containing non-nisq chars
    
    >>> q_str_nn = QuantumString(size = len("hello world"), nisq_char = False)
    >>> q_str_nn[:] = "hello world"
    >>> print(q_str_nn)    
    {'hello world': 1.0}
    
    This requires however considerably more qubits
    
    >>> print(len(q_str.qs.qubits))
    55
    >>> print(len(q_str_nn.qs.qubits))
    88
    
    Similar to it's parent class, the size of a QuantumString does not have to be specified at creation
    
    >>> q_str = QuantumString()
    >>> q_str[:] = "hello world"
    >>> print(q_str)
    {'hello world': 1.0}
    
    **Concatenation**
    
    QuantumStrings provide a number of methods to concatenate:
        
    >>> q_str_0 = QuantumString()
    >>> q_str_1 = QuantumString()
    >>> q_str_2 = QuantumString()
    >>> q_str_0[:] = "hello"
    >>> q_str_1 += " "
    >>> q_str_2[:] = "world"
    >>> q_str_3 = q_str_1 + q_str_2
    >>> q_str_0 += q_str_3
    >>> print(q_str_0)
    {'hello world': 1.0}
    
    Note that these QuantumStrings share memory - ie. if we modify a QuantumChar in one of them, this will potentially affect the others:
        
    >>> from qrisp import h
    >>> h(q_str_2[0][0])
    >>> print(q_str_0)
    {'hello world': 0.5, 'hello xorld': 0.5}
    
    """
    
    def __new__(subtype, size = 0, qs = None, nisq_char = True):
        
        if nisq_char:
            return QuantumArray.__new__(subtype, nisq_init_quantum_char, size,  qs = qs)
        else:
            return QuantumArray.__new__(subtype, init_quantum_char, size,  qs = qs)

    def get_measurement(self, **kwargs):
        mes_result = QuantumArray.get_measurement(self, **kwargs)
        
        return_dic = {}
        
        for (k,v) in mes_result.items():
            return_dic["".join(list(k))] = v
        
        return return_dic
    
    
    def __add__(self, other):
        return np.concatenate((self, other)).view(QuantumString)
    
    def decoder(self, code_int):
        
        res_array = QuantumArray.decoder(self, code_int)
        res_str = ""
        
        for i in range(len(res_array)):
            res_str += res_array[i]
            
        return res_str
    
    def encoder(self, encoding_str):
        
        encoding_array = np.zeros(len(encoding_str), dtype = "object")
        
        for i in range(len(encoding_array)):
            encoding_array[i] = encoding_str[i]
        
        return QuantumArray.encoder(self, encoding_array)
    
    def __setitem__(self, key, value):
        if isinstance(key, int):
            QuantumArray.__setitem__(self, key, value)
            return
        
        QuantumArray.__setitem__(self, key, [ch for ch in value])
    
    
    def __iadd__(self, other):
        
        from qrisp import merge
        
        if isinstance(other, QuantumString):
            self.resize((len(self) + len(other),), refcheck = False)
            merge(self.qs, other.qs)
            for i in range(len(other)):
                np.ndarray.__setitem__(self, i + len(self) - len(other), other[i])
                
                
        elif isinstance(other, str):
            if self.shape_specified:
                self.resize((len(self) + len(other),), refcheck = False)
            else:
                self.set_shape(len(other))
            for i in range(len(other)):
                np.ndarray.__setitem__(self, i + len(self) - len(other), self.qtype.duplicate(qs = self.qs))
                self[i + len(self) - len(other)] = other[i]
                
        
        elif isinstance(other, QuantumChar):
            self.resize((len(self) + 1, ), refcheck = False)
            np.ndarray.__setitem__(self, len(self) - 1, other)
            
            
        return self
        



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


from networkx import DiGraph
import networkx as nx
from qrisp.uncomputation.type_checker import is_permeable, is_qfree
from qrisp.circuit import ControlledOperation, PTControlledOperation, QubitDealloc, QubitAlloc, Instruction, QuantumCircuit
from qrisp.logic_synthesis import LogicSynthGate


class UnqompNode:
    
    def __init__(self, name, instr = None):
        
        self.name = name
        self.targets = []
        self.controls = []
        self.uncomputed_node = None
        
        self.instr = instr
        self.hash = hash(self.name)
    
    def __hash__(self):
        return self.hash
    
    def __str__(self):
        
        if self.instr is None:
            return self.name
        else:
            return str(self.name)
    
    def __repr__(self):
        
        return str(self)


def qc_from_dag(dag):
    
    res_qc = dag.original_qc.clearcopy()
    
    QuantumCircuit.fast_append = True
    
    for node in nx.topological_sort(dag):
        if node.instr:
            res_qc.append(node.instr)

    QuantumCircuit.fast_append = False
    
    return res_qc


def dag_from_qc(qc, remove_init_nodes = False):
    
    dag = DiGraph()
    
    recent_node_dic = {}
    node_counter = {}
    init_nodes = {}
    
    
    for i in range(len(qc.qubits)):
        
        node = UnqompNode("qubit_" + str(i) + "_0")
        
        node.qubit = qc.qubits[i]
        
        dag.add_node(node)
        recent_node_dic[qc.qubits[i]] = node
        node_counter[qc.qubits[i]] = 1
        init_nodes[qc.qubits[i]] = node
    
    
    dag.init_nodes = init_nodes
    
    dealloc_nodes = []
    
    for i in range(len(qc.data)):
        
        instr = qc.data[i]
        
        # if instr.op.name == "barrier":
            # continue
        
        #node = UnqompNode(str(qc.qubits.index(instr.qubits[-1])) + "_" + str(node_counter[instr.qubits[-1]]), instr)
        node = UnqompNode(str(node_counter[instr.qubits[-1]]), instr)
        
        if instr.op.name == "qb_dealloc":
            dealloc_nodes.append(node)
        
        dag.add_node(node)
        for j in range(len(instr.qubits)):
            
            qb = instr.qubits[j]
            
            
            if is_permeable(instr.op, [j]):
                edge_type = "control"
                node.controls.append(qb)
            else:
                edge_type = "target"
            
            if dag.has_edge(recent_node_dic[qb], node) and dag.get_edge_data(recent_node_dic[qb], node)["edge_type"] != "anti_dependency":
                dag.get_edge_data(recent_node_dic[qb], node)["qubits"].append(qb)
            else:
                dag.add_edge(recent_node_dic[qb], node, edge_type = edge_type, qubits = [qb])
            
            successors = list(dag.successors(recent_node_dic[qb]))
            if len(successors) > 1:
                if edge_type == "target":
                    for s in successors:
                        if s == node:
                            continue
                        elif not dag.has_edge(s, node):
                            if dag.get_edge_data(recent_node_dic[qb], s)["edge_type"] == "control":
                                dag.add_edge(s, node, edge_type = "anti_dependency")
            
            if edge_type == "target":
                recent_node_dic[qb] = node
                node_counter[qb] += 1
                node.targets.append(qb)
                
    
    dag.original_qc = qc
    
    if remove_init_nodes:
    #Remove init nodes
        dag.remove_nodes_from(list(init_nodes.values()))
    
    return dag



        
def uncompute_node(dag, uncomp_node, uncomp_qbs, recompute_qubits = []):
    
    
    lin = list(nx.topological_sort(dag))
    lin.reverse()



    a_star_n_list = []
    
    uncomp_node_target_set = set(uncomp_node.targets)
    
    for n in lin:
        if set(uncomp_node.targets).intersection(n.targets):
            
            a_star_n_list.append(n)
            uncomp_node_target_set = uncomp_node_target_set - set(n.targets)
            
            if not len(uncomp_node_target_set):
                break
            
    
    if not is_qfree(uncomp_node.instr.op):
        raise Exception(f"Tried to uncompute non-qfree gate {uncomp_node.instr}")
    
    ctrls = [n for n in dag.predecessors(uncomp_node) if dag.get_edge_data(n, uncomp_node)["edge_type"] == "control"]
    
    for i in range(len(ctrls)):
        if not ctrls[i].uncomputed_node is None:
            ctrls[i] = ctrls[i].uncomputed_node
    
    
    #Replace controlled operations by phase tolerant controlled operations
    if isinstance(uncomp_node.instr.op, ControlledOperation):
        
        op = uncomp_node.instr.op
        
        if op.method == "auto" and 1 < len(op.controls) < 5 or op.method == "gray":
            uncomp_node.instr.op = PTControlledOperation(op.base_operation, num_ctrl_qubits = len(op.controls), ctrl_state = op.ctrl_state)
            
    
    #Replace results of logic synthesis by phase tolerant logic synthesis
    if isinstance(uncomp_node.instr.op, LogicSynthGate):
        if uncomp_node.instr.op.logic_synth_method == "gray":
            
            from qrisp import QuantumVariable
            
            tt = uncomp_node.instr.op.tt
            
            input_qv = QuantumVariable(tt.bit_amount)
            output_qv = QuantumVariable(tt.shape[1], qs = input_qv.qs)
            
            tt.q_synth(input_qv, output_qv, method = "gray_pt")
            
            temp = input_qv.qs.data[-1].op
            temp.name = uncomp_node.instr.op.name
            
            uncomp_node.instr.op = temp
    
    new_instr = uncomp_node.instr.inverse()
    
    reversed_node = UnqompNode(uncomp_node.name + "_*", new_instr)
    reversed_node.controls = uncomp_node.controls
    reversed_node.targets = uncomp_node.targets
    
    uncomp_node.uncomputed_node = reversed_node
    
    dag.add_node(reversed_node)
    
    for i in range(len(ctrls)):
        c = ctrls[i]
        
        control_edge_qubits = dag.get_edge_data(c, uncomp_node)["qubits"]
        
        if set(control_edge_qubits).intersection(recompute_qubits):
            for k in dag.successors(c):
                if dag.get_edge_data(c, k)["edge_type"] == "target" and set(dag.get_edge_data(c, k)["qubits"]).intersection(control_edge_qubits):
                    if k.uncomputed_node is None:
                        return True
                    
                    ctrls[i] = k.uncomputed_node
                    break
            else:
                return True
                
        dag.add_edge(ctrls[i], reversed_node, edge_type = "control", qubits = control_edge_qubits)
    
    
    for a_star_n in a_star_n_list:
        
        dag.add_edge(a_star_n, reversed_node, edge_type = "target", qubits = list(set(uncomp_qbs).intersection(n.targets).intersection(new_instr.qubits)))
        
        for v in dag.successors(a_star_n):
            
            if dag.get_edge_data(a_star_n, v)["edge_type"] == "control":
                dag.add_edge(v, reversed_node, edge_type = "anti_dependency")
    
    
                
    for c in ctrls:
        for v in dag.successors(c):
            if dag.get_edge_data(c, v)["edge_type"] == "target":
                
                target_edge_qubits = dag.get_edge_data(c, v)["qubits"]
                control_edge_qubits = dag.get_edge_data(c, uncomp_node)["qubits"]
                
                if set(target_edge_qubits).intersection(control_edge_qubits):
                    pass
                    dag.add_edge(reversed_node, v, edge_type = "anti_dependency")
                    
    return False
                

def uncompute_qc(qc, uncomp_qbs, recompute_qubits = []):

    qc_new = qc.copy()
    previous_instructions = []
    follow_up_instructions = []
    
    while len(qc_new.data):
        
        instr = qc_new.data[0]
        if set(uncomp_qbs + recompute_qubits).intersection(instr.qubits):
            break
        previous_instructions.append(qc_new.data.pop(0))
    
    while len(qc_new.data):
        
        instr = qc_new.data[-1]
        if set(uncomp_qbs + recompute_qubits).intersection(instr.qubits):
            break
        follow_up_instructions.append(qc_new.data.pop(-1))
        
    follow_up_instructions = follow_up_instructions[::-1]

    dag = dag_from_qc(qc_new)
    
    lin = list(nx.topological_sort(dag))
    lin.reverse()
    
    for i in range(len(lin)):
    # for node in lin:
        
        node = lin[i]
        
        if node.instr is None:
            continue
        
        if not set(node.targets).intersection(uncomp_qbs):
            continue

        if node.instr.op.name in ["barrier"]:#, "qb_alloc", "qb_dealloc"]:
            continue
        
        #This function checks if there is more than one allocation gate in the upcoming
        #chain of nodes. This prevents instructions which were created due to previous recomputations
        #to be included in further recomputations
        if detect_double_alloc(lin[i:], node.instr.qubits):
            continue
        
        if set(node.targets).issubset(uncomp_qbs) and node.instr and node.targets:
            recompute = uncompute_node(dag, node, uncomp_qbs, recompute_qubits)
            if recompute:
                
                return uncompute_qc(qc, uncomp_qbs + list(set(node.controls).intersection(recompute_qubits)), recompute_qubits)
            continue
        
        
        off_target_qubits = list(set(node.targets)-set(uncomp_qbs))
        if off_target_qubits:
            not_uncomputable_qubits = []
            for qb in off_target_qubits:
                if qb.allocated:
                    not_uncomputable_qubits.append(qb)
            
            if not_uncomputable_qubits:
                raise Exception(f"Uncomputation failed because gate \"{node.instr.op.name}\" needs to be uncomputed but is also targeting qubits {not_uncomputable_qubits} which are not up for uncomputation")
            else:
                return uncompute_qc(qc, uncomp_qbs + off_target_qubits, recompute_qubits = recompute_qubits)
    
    
    try:
        uncomputed_qc = qc_from_dag(dag)
    except nx.NetworkXUnfeasible:
        raise Exception("Cycle detected in DAG during uncomputation")
    
    
    uncomputed_qc.data = previous_instructions + uncomputed_qc.data + follow_up_instructions
    
    return uncomputed_qc

def detect_double_alloc(lin, qubits):
    
    qb_allocs = []
    for node in lin:
        if not node.instr:
            continue
        if node.instr.op.name == "qb_alloc" and node.instr.qubits[0] in qubits:
            if node.instr.qubits[0] in qb_allocs:
                return True
            else:
                qb_allocs.append(node.instr.qubits[0])
                
    return False





"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


import threading
import os

import connexion
import six

import qrisp.interface.openapi_interface.codegen.server.openapi_server as openapi_server
from openapi_server.models.base_model_ import Model
from openapi_server.models.inline_object import InlineObject  # noqa: E501
from openapi_server.models.backend_status import BackendStatus

from connexion.apps.flask_app import FlaskJSONEncoder

class JSONEncoder(FlaskJSONEncoder):
    include_nulls = False

    def default(self, o):
        if isinstance(o, Model):
            dikt = {}
            for attr, _ in six.iteritems(o.openapi_types):
                value = getattr(o, attr)
                if value is None and not self.include_nulls:
                    continue
                attr = o.attribute_map[attr]
                dikt[attr] = value
            return dikt
        return FlaskJSONEncoder.default(self, o)


#Returns the hosts ip
def get_ip():
    import socket
    s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
    s.settimeout(0)
    try:
        # doesn't even have to be reachable
        s.connect(('10.255.255.255', 1))
        IP = s.getsockname()[0]
    except Exception:
        IP = '127.0.0.1'
    finally:
        s.close()
    return IP



#This class describes a Backend Server
class BackendServer:
    
    
    #For the constructor we specify an function that runs the circuit, a socket ip adress, a port and the.
    #Note that the name is supposed to be unique, that is if this python instance already
    #runs a threat with that name, no new server is started
    def __init__(self, run_func, socket_ip_address = None, port = None, name = "generic_quantum_backend_server", ping_func = None, online_status = True):
        
        self.app = connexion.App(__name__, specification_dir='./codegen/server/openapi_server/openapi/')
        self.app.app.json_encoder = JSONEncoder


        @self.app.route("/run", methods=["POST"])
        def run():
            inline_object = InlineObject.from_dict(connexion.request.get_json())
            qc = inline_object.qc
            shots = inline_object.shots
            token = inline_object.token
            
            return run_func(qc, shots, token)
        
        if ping_func != None:
            @self.app.route("/ping", methods=["GET"])
            def ping():
                res = ping_func()
                res.online = pass_online_status_bool_by_reference()
                return res.to_dict()
        
            
        self.online_status = online_status
        pass_online_status_bool_by_reference = lambda : self.online_status
        
        self.name = name
        
        if port == None:
            port = 8080
        
        if socket_ip_address == None:
            socket_ip_address = get_ip()
            
        self.socket_ip_address = socket_ip_address
        self.port = port
    
    #Starts the server
    def start(self):
        
        
        #self.app.run(port=self.port)
        
        #import logging
        #log = logging.getLogger('werkzeug')
        #log.setLevel(logging.ERROR)
        #Create thread
        from waitress import serve
        
        def wrapper():
            serve(self.app, host = self.socket_ip_address, port = self.port)
            
        thr = threading.Thread(target = wrapper)
        thr.setDaemon(True)
        
        #Start the thread
        thr.start()



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# Created by ann81984 at 28.06.2022
from qrisp import QuantumDictionary, QuantumFloat, custom_qv,  h, multi_measurement, QuantumVariable
#This example demonstrates the use of a QuantumDictionary with flexible types

def test_quantum_dictionary():
    #Create a QuantumDictionary
    test_qd = QuantumDictionary()

    #Fill in some entries (works precisely like normal Python dictionaries)
    test_qd["hello"] = "hallo"
    test_qd["world"] = "welt"

    #Create Key QuantumVariable (the constructor takes a
    #list of values, which the QuantumVariable is supposed to represent)
    key_qv = custom_qv(["hello", "world"])

    #Bring into superposition
    h(key_qv)

    print(key_qv)
    assert str(key_qv) == "{'hello': 0.5, 'world': 0.5}"
    #Yields {'hello': 0.5, 'world': 0.5}

    #Use QuantumDictionary to "dereference" the values into a new QuantumVariable
    value_qv = test_qd[key_qv]

    #Evaluate values
    print(value_qv)
    assert str(value_qv) == "{'hallo': 0.5, 'welt': 0.5}"
    #Yields {'hallo': 0.5, 'welt': 0.5}


    #Now we demonstrate how a QuantumDictionary with a specified type can be created
    #The return values from dereferencing this QuantumDictionary are now of the type QuantumFloat(4, -1)
    float_qd = QuantumDictionary(return_type = QuantumFloat(4,-1))

    #Encode some values (performed as with usual Python dictionaries)
    float_qd.update({"hallo" : 1.5, "welt" : 0.5})

    #Dereference
    value_qf = float_qd[value_qv]

    #Results are indeed QuantumFloat
    print(type(value_qf))
    assert type(value_qf).__name__ == 'QuantumFloat'

    #Usual arithmetic is possible
    value_qf += 2.5

    #Evaluate results
    assert value_qf.get_measurement() == {3.0: 0.5, 4.0: 0.5}
    #Yields {4.0: 0.5, 3.0: 0.5}
    
    
    test_qd[(1,2)] = 73
    test_qd[(0,2)] = 37
    qf1 = QuantumFloat(1)
    qf2 = QuantumFloat(2)
    h(qf1)
    qf2[:] = 2
    res = test_qd[(qf1, qf2)]
    assert multi_measurement([qf1, qf2, res]) == {(0, 2, 37): 0.5, (1, 2, 73): 0.5}
    
    
    qtype = QuantumFloat(4, -2, signed = True)

    float_qd = QuantumDictionary(return_type = qtype)

    float_qd["hello"] = 0.5
    float_qd["world"] = -1
    key_qv = QuantumVariable.custom([1, 42, "hello", "world"])
    key_qv[:] = {"hello" : 1, "world" : 1}


    qf = qtype.duplicate()
    float_qd.load(key_qv, qf)

    qf2 = qtype.duplicate()
    float_qd.load(key_qv, qf2, synth_method = "pprm")
    
    assert qf2.get_measurement() == {0.5: 0.5, -1.0: 0.5}
    
    


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


import copy
import weakref

import numpy as np
import matplotlib.pyplot as plt

from qrisp.circuit import transpile
from qrisp.core.compilation import qompiler

class QuantumVariable:
    """
    The QuantumVariable is the quantum equivalent of a regular variable in classical programming languages. All :ref:`quantum types <QuantumTypes>` inherit from this class.
    The QuantumVariable allows many automizations and quality of life improvements such as hidden qubit management, de/encoding to human readable labels or typing.
    
    Each QuantumVariable is registered in a :ref:`QuantumSession`. It can be accessed using the ``.qs`` attribute:
        
    >>> from qrisp import QuantumVariable
    >>> example_qv = QuantumVariable(3)
    >>> quantum_session = example_qv.qs

    The qubits of the QuantumVariable are stored as a list in the ``.reg`` attribute

    >>> qubits = example_qv.reg
    
    To quickly access the qubits of a given variable, we use the [ ] operator:
    
    >>> qubit_2 = example_qv[2]
    
    We can find out about the amount of qubits in the QuantumVariable with the ``.size`` attribute
    
    >>> example_qv.size
    3
    
    **Naming**
    
    QuantumVariables can be given names to identify them independently of their naming as Python objects. 
    
    >>> example_qv_2 = QuantumVariable(3, name = "alice")
    >>> example_qv_2.name
    'alice'
    
    If not explicitely specified during construction, a name is determined automatically. Qrisp will try to infer the name of the Python variable and if that fails, a generic name is given.
    
    >>> example_qv.name
    'example_qv'
    
    In order to keep the generated quantum circuits comprehensive, the qubits are named after their containing QuantumVariable with an extra number, which indicates their index.
    
    >>> from qrisp import cx
    >>> cx(example_qv, example_qv_2)
    >>> print(example_qv.qs)
    QuantumCircuit:
    ---------------
    example_qv.0: ──■────────────
                    │            
    example_qv.1: ──┼────■───────
                    │    │       
    example_qv.2: ──┼────┼────■──
                  ┌─┴─┐  │    │  
         alice.0: ┤ X ├──┼────┼──
                  └───┘┌─┴─┐  │  
         alice.1: ─────┤ X ├──┼──
                       └───┘┌─┴─┐
         alice.2: ──────────┤ X ├
                            └───┘
    Live QuantumVariables:
    ----------------------
    QuantumVariable example_qv
    QuantumVariable alice    
    
    QuantumSessions can only contain uniquely named QuantumVariables. If two QuantumSessions are :ref:`merged <SessionMerging>` containing identically named QuantumVariables, the more recently created QuantumVariable will be renamed:
    
    ::
        
        from qrisp import QuantumFloat
        
        s = QuantumFloat(5)
        
        for i in range(4):
            temp = QuantumFloat(4)
            temp[:] = 2**i
            s += temp
            
    >>> print(s.qs)
    QuantumCircuit:
    ---------------
                   ┌───────────┐┌───────────┐┌───────────┐┌───────────┐
         s.0: ─────┤0          ├┤0          ├┤0          ├┤0          ├
                   │           ││           ││           ││           │
         s.1: ─────┤1          ├┤1          ├┤1          ├┤1          ├
                   │           ││           ││           ││           │
         s.2: ─────┤2          ├┤2          ├┤2          ├┤2          ├
                   │           ││           ││           ││           │
         s.3: ─────┤3          ├┤3          ├┤3          ├┤3          ├
                   │           ││           ││           ││           │
         s.4: ─────┤4 __iadd__ ├┤4          ├┤4          ├┤4          ├
              ┌───┐│           ││           ││           ││           │
      temp.0: ┤ X ├┤5          ├┤           ├┤           ├┤           ├
              └───┘│           ││           ││           ││           │
      temp.1: ─────┤6          ├┤  __iadd__ ├┤           ├┤           ├
                   │           ││           ││           ││           │
      temp.2: ─────┤7          ├┤           ├┤           ├┤           ├
                   │           ││           ││           ││           │
      temp.3: ─────┤8          ├┤           ├┤  __iadd__ ├┤           ├
                   └───────────┘│           ││           ││           │
    temp_1.0: ──────────────────┤5          ├┤           ├┤           ├
              ┌───┐             │           ││           ││           │
    temp_1.1: ┤ X ├─────────────┤6          ├┤           ├┤  __iadd__ ├
              └───┘             │           ││           ││           │
    temp_1.2: ──────────────────┤7          ├┤           ├┤           ├
                                │           ││           ││           │
    temp_1.3: ──────────────────┤8          ├┤           ├┤           ├
                                └───────────┘│           ││           │
    temp_2.0: ───────────────────────────────┤5          ├┤           ├
                                             │           ││           │
    temp_2.1: ───────────────────────────────┤6          ├┤           ├
              ┌───┐                          │           ││           │
    temp_2.2: ┤ X ├──────────────────────────┤7          ├┤           ├
              └───┘                          │           ││           │
    temp_2.3: ───────────────────────────────┤8          ├┤           ├
                                             └───────────┘│           │
    temp_3.0: ────────────────────────────────────────────┤5          ├
                                                          │           │
    temp_3.1: ────────────────────────────────────────────┤6          ├
                                                          │           │
    temp_3.2: ────────────────────────────────────────────┤7          ├
              ┌───┐                                       │           │
    temp_3.3: ┤ X ├───────────────────────────────────────┤8          ├
              └───┘                                       └───────────┘
    Live QuantumVariables:
    ----------------------
    QuantumFloat s
    QuantumFloat temp
    QuantumFloat temp_1
    QuantumFloat temp_2
    QuantumFloat temp_3
        
    Renaming does not happen for names given through the ``name`` keyword, unless the name ends with a ``*``.
    
    >>> example_qv_3 = QuantumVariable(3, name = "alice")
    >>> cx(example_qv, example_qv_3)
    Exception: Tried to merge QuantumSession containing identically named QuantumVariables
    >>> example_qv_4 = QuantumVariable(3, name = "alice*")    
    >>> cx(example_qv, example_qv_4)
    >>> example_qv_4.name
    'alice_1'
    
    
    Examples
    --------
    
    Writing a function that brings an arbitrary QuantumVariable into a GHZ state
    
    ::
        
        from qrisp import QuantumVariable, h, cx
        def GHZ(qv):
             h(qv[0])
             for i in range(1, qv.size): 
                 cx(qv[0], qv[i])
        
    Evaluation:
    
    >>> qv = QuantumVariable(5)
    >>> GHZ(qv)
    >>> print(qv)
    {'00000': 0.5, '11111': 0.5}
    
    
    """
    
    
    
    live_qvs = []
    creation_counter = np.zeros(1)
    name_tracker = {}
    
    def __init__(self, size, qs = None, name = None):
        """
        Constructs a QuantumVariable - possibly with a given name or in a given QuantumSession.

        Parameters
        ----------
        size : int
            The amount of qubits this QuantumVariable contains.
        qs : QuantumSession, optional
            A QuantumSession object, where the QuantumVariable is supposed to be registerd. The default is None.
        name : string, optional
            A name which uniquely identifies the QuantumVariable. If ended with a \*, name is allowed to be updated if a naming collision arises. By default, Qrisp will try to infer the name of the Python variable - otherwise a generic name is given.

        """

        if size < 0:
            raise Exception(f"Tried to create QuantumVariable with invalid qubit amount {size}")
        

        #Store quantum session        
        from qrisp.core import QuantumSession, merge_sessions
        self.qs = QuantumSession()
        
        if not qs is None:
            merge_sessions(self.qs, qs)
            

        self.size = int(size)
        
        
        self.user_given_name = False
        #If name is given, register variable in session manager
        if not name is None:
            self.user_given_name = True
            
            if name[-1] == "*":
                name = name[:-1]
                self.user_given_name = False
                
                try:
                    self.name = name
                    self.qs.register_qv(self)
                    
                except RuntimeError:
                    i = 0
                    while True:
                        try:
                            self.name = name + "_" + str(i)
                            self.qs.register_qv(self)
                        except RuntimeError:
                            i += 1
                            continue
                        break
                    
            else:
                self.name = name
                self.qs.register_qv(self)
        
        #Otherwise try to infer from code inspection
        else:
            from qrisp.misc import find_calling_line
            
            if type(self) is QuantumVariable:
                line = find_calling_line(1)
            else:
                line = find_calling_line(2)
            split_line = line.split("=")
            name_found = False
            
            if len(split_line) >= 2:
                python_var_name = split_line[0]
                
                if split_line[1].replace(" ", "")[:7] == "Quantum":
                    
                    python_var_name = python_var_name.split(" ")[0]
                    python_var_name = python_var_name.split(" ")[-1]
                    
                    # name = self.get_unique_name(python_var_name)
                    name = python_var_name
                    
                    self.name = name
                    self.qs.register_qv(self)    
                    name_found = True

            #If this didn't work, generate a generic, unique name
            if not name_found:
                while True:
                    try:
                        self.name = self.get_unique_name()
                        self.qs.register_qv(self)
                        break
                    except RuntimeError:
                        pass
        
        import weakref
        #This attribute tracks the created QuantumVariables for the 
        #auto_uncompute decorator
        #We use weak references as some qrisp modules rely on reference counting
        QuantumVariable.live_qvs.append(weakref.ref(self))
        self.creation_time = int(self.creation_counter[0])
        self.creation_counter += 1
    
    
    def delete(self, verify = False, recompute = False):
        """
        This method is for deleting a QuantumVariable and thus freeing up and resetting the used qubits.
        
        Note that this method has a different function than the destructor. Calling this method will tell the QuantumSession to mark the used qubits as free and apply a reset gate.
        If set to True, the keyword verify will cause a simulation to check, wether the deleted qubits are in the $\ket{0}$ state prior to resetting. This is helpfull during debugging, as it indicates wether the uncomputation of this QuantumVariable was successfull.
        
        After deletion, the QuantumVariable object is basically unchanged but an error will be raised if further operations on the deleted qubits are attempted.

        Parameters
        ----------
        verify : bool, optional
            If this bool is set to True, Qrisp will verify that the deleted qubits are indeed in the $\ket{0}$ state. The default is ``False``.
        recompute : bool, optional
            If set to ``True``, this QuantumVariable can be recomputed if it is required for the uncomputation of another QuantumVariable. For more information on the (dis)advantages check :ref:`recomputation <recomputation>`. The default is ``False``.

        Raises
        ------
        Exception
            Tried to delete qubits not in \|0> state.

        
        Examples
        --------
        
        We create a QuantumVariable, execute some gates and try to delete with verify = True
            
        >>> from qrisp import QuantumVariable, x, h
        >>> qv = QuantumVariable(2)
        >>> x(qv[0])
        >>> h(qv[1])
        >>> qv.delete(verify = True)
        Exception: Tried to delete qubits not in |0> state.
        
        We now (manually) uncompute the gates
        
        >>> x(qv[0])
        >>> h(qv[1])
        >>> qv.delete(verify = True)
        >>> qv.is_deleted()
        True
        >>> x(qv[0])
        Exception: Tried to perform operation x on unallocated qubit qv_1.0.
        
        """
        
        if self.is_deleted():
            return
        
        self.qs.delete_qv(self, verify)
        
        i = 0
        while i < len(QuantumVariable.live_qvs):
            if QuantumVariable.live_qvs[i]() is None:
                QuantumVariable.live_qvs.pop(i)
                continue
            
            if QuantumVariable.live_qvs[i]().name == self.name:
                QuantumVariable.live_qvs.pop(i)
                break
            
            i += 1
        
        if recompute:
            for qb in self.reg:
                qb.recompute = True
        
            
    def is_deleted(self):
        for qb in self.reg:
            if qb.allocated:
                return False
        else:
            return True
        
    
    def duplicate(self, name = None, qs = None, init = False):
        """
        Duplicates the QuantumVariable in the sense that a new QuantumVariable is created with same type and parameters but initialized in the $\ket{0}$ state.
        
        Parameters
        ----------
        name : string, optional
            A unique name to identify that QuantumVariable. If not given, a name will be generated.
        qs : QuantumSession, optional
            A QuantumSession, where the result should be registered. If not given, a new QuantumSession will be generated.
        init : bool, optional
            If set to True, the :meth:`init_from <qrisp.QuantumVariable.init_from>` method of the result will be called on self. The default is False.

        Returns
        -------
        duplicate : Type of self
            The duplicated QuantumVariable.

        Examples
        --------
        
        We create a QuantumFloat and duplicate:
        
        >>> from qrisp import QuantumFloat
        >>> qf_0 = QuantumFloat(4, signed = False)
        >>> qf_1 = qf_0.duplicate()
        >>> type(qf_1)
        qrisp.qtypes.quantum_float.QuantumFloat
        >>> qf_1.size
        4
        
        """
        
        duplicate = copy.copy(self)
        
        from qrisp.core import QuantumSession
        new_qs = QuantumSession()

        #Register duplicate variable in session manager
        
        
        if not name is None:
            duplicate.name = name
            new_qs.register_qv(duplicate)
            duplicate.user_given_name = True
            
        else:
            duplicate.user_given_name = False
            
            try:
                duplicate.name = self.name
                new_qs.register_qv(duplicate)
            except NameError:
                i = 0
                while True:
                    try:
                        duplicate.name = self.name + "_" + str(i)
                        new_qs.register_qv(duplicate)
                        break
                    except NameError:
                        pass
                    i += 1

        from qrisp import merge
        
        
        
        duplicate.qs = new_qs
        
        if not qs is None:
            merge(qs, new_qs)
        
        #This attribute tracks the created QuantumVariables for the 
        #auto_uncompute decorator
        #We use weak references as some qrisp modules rely on reference counting
        QuantumVariable.live_qvs.append(weakref.ref(duplicate))
        duplicate.creation_time = int(self.creation_counter[0])
        duplicate.creation_counter += 1
        
        
        if init:
            duplicate.init_from(self)

        

        
        return duplicate        
    
    def decoder(self, i):
        """
        The decoder method specifies how a QuantumVariable turns the outcomes of measurements into human readable values. It recieves an integer ``i`` and returns a human readable value.
        
        This method is supposed to be overloaded when defining new :ref:`quantum types <QuantumTypes>`.
        
        
        Parameters
        ----------
        i : int
            Integer representing the outcome of a measurement of the qubits of this QuantumVariable.

        Returns
        -------
        
            A human readable value. Has to be hashable.

        Examples
        --------
        
        We create a QuantumFloat and inspect it's decoder:
        
        >>> from qrisp import QuantumFloat
        >>> qf = QuantumFloat(3, -1, signed = False)
        >>> print(qf.decoder(1))
        0.5
        
        This implies that if the 3 qubits of this QuantumFloat are measured in state 001, this outcome corresponds to the value 0.5.
        """
        
        from qrisp.misc import bin_rep
        return bin_rep(i, self.size)[::-1]
    
    
    def encoder(self, value):
        """
        The encoder reverses the decoder, it turns human readable values into integers.
        
        If not overloaded, the encoder will perform a linear search on decoder inputs to match the given value.

        Parameters
        ----------
        label : 
            A human readable value.

        Raises
        ------
        Exception
            Unknown input value.

        Returns
        -------
        i : int
            The integer encoding the given value.

        Examples
        --------

        We create a QuantumChar and inspect it's encoder:
        
        >>> from qrisp import QuantumChar
        >>> q_ch = QuantumChar()
        >>> print(q_ch.encoder("f"))
        5
        
        This implies that if the 5 qubits of this QuantumChar are measured to ``5 = 00101``, the out come will be displayed as f.
        
        """
        for i in range(2**self.size):
            if self.decoder(i) == value:
                return i
        
        raise Exception("Value " + str(value) + " not supported by encoder.")
        

    def encode(self, value):
        """
        The encode method allows to quickly bring a QuantumVariable in a desired computational basis state.
        
        A shorthand for this method is given by the ``[:]`` operator.
        
        Note that the qubits to initialize have to be fresh (ie. no operations performed on them).
        
        Parameters
        ----------
        value : 
            A value supported by the encoder.

        Returns
        -------
        None.

        Examples
        --------
        
        We create two quantum floats and encode the value 2.5. For one of them, we perform an x gate onto the corresponding qubits, resulting in an error.
        
        >>> from qrisp import QuantumFloat, x
        >>> qf_0 = QuantumFloat(3, -1, signed = False)
        >>> qf_1 = QuantumFloat(3, -1, signed = False)
        >>> x(qf_0)
        >>> qf_0.encode(2.5)
        Exception: Tried to initialize qubits which are not fresh anymore.
        >>> qf_1[:] = 2.5
        >>> print(qf_1)
        {2.5: 1.0}
        
        """
        
        from qrisp.misc import int_encoder, check_if_fresh
        
        if not check_if_fresh(self.reg, self.qs):
            raise Exception("Tried to initialize qubits which are not fresh anymore.")
        
        
        int_encoder(self, self.encoder(value))
        
        

    def init_state(self, state_dic):
        """
        The ``init_state`` method allows the initialization of arbitrary quantum states. It recieves a dictionary of the type 
        
        **{value : complex number}**
        
        and initializes the **normalized** state. Amplitudes not specified are assumed to be zero.
    
        Note that the state initialization algorithm requires it's qubits to be in state $\ket{0}$.
        
        A shorthand for this method is the ``[:]`` operator, when handed the corresponding dictionary
        
        Parameters
        ----------
        state_dic : dict
            Dictionary describing the wave function to be initialized.
        
        Raises
        ------
        Exception
            Tried to initialize qubits which are not fresh anymore.
        
        Examples
        --------
        
        We create a QuantumFloat and encode the state 
        
        .. math::
            
            \ket{\psi} = \\sqrt{\\frac{1}{3}} \ket{0.5} + i\\sqrt{\\frac{2}{3}} \ket{2}
        
        >>> from qrisp import QuantumFloat
        >>> qf = QuantumFloat(3, -1)
        
        We can now use either
        
        >>> qf.init_state({0.5: (1/3)**0.5, 2.0 : 1j*(2/3)**0.5})
        
        or:
        
        >>> qf[:] = {0.5: (1/3)**0.5, 2.0 : 1j*(2/3)**0.5}
        
        To acquire the expected result        
                
        >>> print(qf)
        {2.0: 0.6667, 0.5: 0.3333}
        
        """
        
        from qrisp.misc import check_if_fresh
        if not check_if_fresh(self.reg, self.qs):
            raise Exception("Tried to initialize qubits which are not fresh anymore.")
        
        
        from qrisp import init_state
        
        target_array = np.zeros(2**self.size, dtype = np.complex128)
        
        for key in state_dic.keys():
            target_array[self.encoder(key)] = state_dic[key]
        
        
        target_array = target_array/np.vdot(target_array, target_array)**0.5
        
        init_state(self, target_array)

    def append(self, operation):
        self.qs.append(operation, self)
    
    def extend(self, amount, position = -1):
        """
        This method is used to add more qubits to the QuantumVariable. Using the position keyword it is possible to specify the position where the qubits should be added. By default, the qubits are added at the end.

        Parameters
        ----------
        amount : int
            The amount of qubits to add.
        position : int, optional
            The position of where to add the qubits. By default, qubits are added at the end.
            st of qubits which are to be added to the QuantumVariable. The default is None.

        Raises
        ------
        Exception
            Missmatch between proposed qubits and amount integer.

        Returns
        -------
        None.

        Examples
        --------
        
        We create a QuantumVariable and extend it with some extra qubits.
            
        >>> from qrisp import QuantumVariable
        >>> qv = QuantumVariable(3)
        >>> print(qv.reg)
        [Qubit(qv.0), Qubit(qv.1), Qubit(qv.2)]
        >>> qv.extend(3)
        >>> print(qv.reg)
        [Qubit(qv.0), Qubit(qv.1), Qubit(qv.2), Qubit(qv.6), Qubit(qv.6), Qubit(qv.6)]        

        """
        
        if position == -1:
            position = self.size
        
        insertion_qubits = self.qs.request_qubits(amount)
        
        for i in range(amount):
            insertion_qubits[i].identifier = self.name + "." + str(self.size)
            self.reg.insert(position + i, insertion_qubits[i])
            self.size += 1
    
    def reduce(self, qubits, verify = False):
        """
        Reduces the qubit count of the QuantumVariable by removing a specified set of qubits.

        Parameters
        ----------
        qubits : list
            The qubits to remove from the QuantumVariable.
        
        verify : bool
            Boolean value which indicates wether Qrisp should verify that the reduced qubits are in the $\ket{0}$ state.

        Raises
        ------
        Exception
            Qubits not present in QuantumVariable.
            
        Exception
            Verification that the given qubits are in $\ket{0}$ state failed.

        Examples
        --------
        
        We create a QuantumVariable with 5 qubits and remove the first 2
        
        >>> from qrisp import QuantumVariable
        >>> qv = QuantumVariable(5)
        >>> print(qv.reg)
        [Qubit(qv.0), Qubit(qv.1), Qubit(qv.2), Qubit(qv.3), Qubit(qv.4)]
        >>> qv.reduce(qv[:2])
        >>> print(qv.reg)
        [Qubit(qv.2), Qubit(qv.3), Qubit(qv.4)]

        """
        
        
        try:
            len(qubits)
        except TypeError:
            qubits = [qubits]
        
        if not set(qubits).issubset(self.reg):
            raise Exception("Tried to reduce QuantumVariable by invalid qubits")


        #Find Qubits to be cleared        
        for i in range(len(qubits)):
            for j in range(self.size):
                if self.reg[j] == qubits[i]:
                    self.reg[j].identifier = "qb_" + str(self.qs.qubit_index_counter)
                    self.qs.qubit_index_counter += 1
                    self.reg.pop(j)
                    break
        
        self.qs.clear_qubits(qubits, verify)
        #Adjust variable size
        self.size -= len(qubits)        
    
    
    def get_measurement(self, plot = False, backend = None, shots = 10000, compile = True, compilation_kwargs = {}, subs_dic = {}, circuit_preprocessor = None, filename = None):
        r"""
        Method for quick access to the measurement results of the state of the variable. This method returns a dictionary of the type {value : p} where p indicates the probability with which that value is measured.
        

        Parameters
        ----------
        plot : Bool, optional
            Plots the measurement results as a historgram. The default is False.
        backend : BackendClient, optional
            The backend on which to evaluate the quantum circuit. The default can be specified in the file default_backend.py.
        shots : integer, optional
            The amount of shots to evaluate the circuit. The default is 10000.
        compile : bool, optional
            Boolean indicating if the .compile method of the underlying QuantumSession should be called before. The default is True.
        compilation_kwargs  : dict, optional
            Keyword arguments for the compile method. For more details check :meth:`QuantumSession.compile <qrisp.QuantumSession.compile>`. The default is ``{}``.
        subs_dic : dict, optional
            A dictionary of Sympy symbols and floats to specify parameters in the case of a circuit with unspecified, :ref:`abstract parameters<QuantumCircuit>`. The default is {}.
        circuit_preprocessor : Python function, optional
            A function which recieves a QuantumCircuit and returns one, which is applied after compilation and parameter substitution. The default is None.
        filename : string, optional
            The location of where to save a generated plot. The default is None.

        Raises
        ------
        Exception
            If the containing QuantumSession is in a quantum environment, it is not possible to execute measurements.

        Returns
        -------
        dict
            A dictionary of values and their corresponding measurement probabilities.

        Examples
        --------
        
        We create an integer :ref:`QuantumFloat`, encode the value 1 and bring the qubit with significance 2 in superposition. We utilize the Qiskit transpiler by transpiling into the gate set $\{\text{CX}, \text{U}\}$
        
        >>> from qrisp import QuantumFloat, h
        >>> qf = QuantumFloat(3,-1)
        >>> qf[:] = 1
        >>> h(qf[2])
        >>> mes_results = qf.get_measurement(transpilation_kwargs = {"basis_gates" : ["cx", "u"]})
        >>> print(mes_results)
        {1.0: 0.5, 3.0: 0.5}
        """
        
        if backend == None:
            if self.qs.backend == None:
                from qrisp.default_backend import def_backend
                backend = def_backend
            else:
                backend = self.qs.backend
            
            
        if len(self.qs.env_stack) != 0:
            raise Exception("Tried to get measurement within open environment")
        
        if self.is_deleted():
            raise Exception("Tried to get measurement from deleted QuantumVariable")
        
        if self.size == 0:
            return {"" : 1.0}
        

        #Bind parameters
        if subs_dic:
            qc = self.qs.bind_parameters(subs_dic)
            
        else:
            if compile:
                qc = qompiler(self.qs, intended_measurements = self.reg, **compilation_kwargs)
            else:
                qc = self.qs.copy()
                
        #Copy circuit in over to prevent modification
        from qrisp.quantum_network import QuantumNetworkClient
        
        if isinstance(backend, QuantumNetworkClient):
            self.qs.data = []
            shots = 1
                
        #Execute user specified circuit_preprocessor
        if circuit_preprocessor != None:
            qc = circuit_preprocessor(qc)

        qc = qc.transpile()
        
        from qrisp.misc import get_measurement_from_qc
        
        counts = get_measurement_from_qc(qc, self.reg, backend, shots)
            
        #Insert outcome labels (if available and hashable)
        try:
            new_counts_dic = {}
            
            sorted_keys = list(counts.keys())
            sorted_keys.sort()
            
            for key in sorted_keys:
                new_counts_dic[self.decoder(key)] = counts[key]
            
            counts = new_counts_dic
            
            #Sort keys
            sorted_key_list = list(counts.keys())
            sorted_key_list.sort(key = lambda x : -counts[x])
            counts = {key : counts[key] for key in sorted_key_list}

            
        except TypeError:
            counts_tuple_list = []
            
            for key in counts.keys():
                counts_tuple_list.append((key, counts[key]))
            
            counts = counts_tuple_list
        
            counts.sorted(key = lambda x : x[1])
        
            
        if plot:
            outcome_labels = []
            for i in range(2**self.size):
                
                temp = self.decoder(i)
                try:
                    hash(temp)
                except TypeError:
                    raise Exception("Outcome value " + str(self.decoder(i)) + " is not hashable")
                    
                outcome_labels.append(temp)

            
            
            if outcome_labels[0] == int(self.size)*"0":
                outcome_labels = list(range(2**self.size))
            plot_histogram(outcome_labels, counts, filename)
            plt.show()
        
        
        #Return dictionary of measurement results
        return counts
    
    def most_likely(self, **kwargs):
        """
        Performs a measurement and returns the most likely outcome.

        Parameters
        ----------
        **kwargs : Keyword arguments for the get_measurement call.
        
        Examples
        --------
        
        >>> from qrisp import QuantumFloat, ry
        >>> import numpy as np
        >>> qf = QuantumFloat(3)
        >>> ry(np.pi*9/8, qf[0])
        >>> print(qf)
        {1: 0.9619, 0: 0.0381}
        >>> qf.most_likely()
        1
    
        """
        
        return list(self.get_measurement())[0]
    
    def __getitem__(self, key):
        return self.reg[key]
    
    def __str__(self):
        return str(self.get_measurement())
    
    def __repr__(self):
        return "<" + str(type(self)).split(".")[-1][:-2] + " '" + self.name + "'>"
        return str(type(self)).split(".")[-1][:-2] + "(name = " + self.name + ")"
        return str(self)
    
    def __len__(self):
        return self.size
    
    #Overload equality operator to use python syntax for if environments? Not sure if the possible user confusion is worth it
    def __eq__(self, other):
        
        from qrisp.environments import q_eq
        return q_eq(self, other)
        
    def __hash__(self):
        return id(self)

    def __setitem__(self, key, value):
        if key != slice(None, None, None):
            raise Exception("Tried to encode value into QuantumVariable using non-trivial slicing.")
        
        if isinstance(type(value), type(None)):
            return
        
        if isinstance(value, dict):
            self.init_state(value)
            return
        
        if isinstance(value, QuantumVariable):
            self.init_from(value)
            return
        
        self.encode(value)
        

        
    def app_phase_function(self, phi):
        """
        Applies a previously specified phase function to each computational basis state of the QuantumVariable using Gray-Synthesis.
        
        For a given phase function $\phi(x)$ and a QuantumVariable in state $\ket{\psi} = \sum_{x \in \\text{Labels}} a_x \ket{x}$  this method acts as:
            
        .. math::
        
            U_{\phi} \sum_{x \in \\text{Labels}} a_x \ket{x} =  \sum_{x \in \\text{Labels}} \\text{exp}(i\phi(x)) a_x \ket{x}



        Parameters
        ----------
        phi : Python function
            A Python function which turns the labels of the QuantumVariable into floats.

        Examples
        --------
        
        We create a QuantumFloat and encode the k-th basis state of the Fourier basis. Finally we will apply an inverse Fourier transformation to measure k in the computational basis.
        
        >>> import numpy as np
        >>> from qrisp import QuantumFloat, h, QFT
        >>> n = 5
        >>> qf = QuantumFloat(n, signed = False)
        >>> h(qf)
        
        After this, qf is in the state
        
        .. math::
            
            \\ket{\\text{qf}} = \\frac{1}{\sqrt{2^n}} \sum_{x = 0}^{2^n} \ket{x}
        
        We specify phi
        
        >>> k = 4
        >>> def phi(x):
        >>>     return 2*np.pi*x*k/2**n
        
        And apply phi as a phase function
        
        >>> qf.app_phase_function(phi)
        
        qf is now in the state
        
        .. math::
            
            \\ket{\\text{qf}} = \\frac{1}{\sqrt{2^n}} \sum_{x = 0}^{2^n} \\text{exp}\\left( \\frac{2\pi ikx}{2^n}\\right) \ket{x}
        
        
        Finally we apply the inverse Fourier transformation and measure:
        
        >>> QFT(qf, inv = True)
        >>> print(qf)
        {4: 1.0}
        

        """        

            
        from qrisp.misc import app_phase_function
        
        app_phase_function([self], phi)
        
    def uncompute(self, do_it = True, recompute = False):
        """
        Method for automatic uncomputation. Uses a generalized form of `this algorithm <https://dl.acm.org/doi/10.1145/3453483.3454040>`_.
        
        For more information check the :ref:`uncomputation documentation<uncomputation>`.

        Parameters
        ----------
        do_it : bool, optional
            If set to False, this variable will be appended to the uncomputation stack of it's QuantumSession and uncomputed once an uncompute call with ``do_it = True`` is performed. The default is True.
        recompute : bool, optional
            If set to True, this QuantumVariable will be uncomputed but temporarily recomputed, if it is required for the uncomputation of another QuantumVariable. For more information check :ref:`recomputation <recomputation>`. The default is False.

        Examples
        --------
        
        We create two QuantumVariables, apply some gates and perform automatic uncomputation
        
        >>> from qrisp import QuantumVariable, x, cx, h, p, mcx
        >>> a = QuantumVariable(3)
        >>> b = QuantumVariable(2)
        >>> mcx(a, b[0])
        >>> h(a[:2])
        >>> x(b[0])
        >>> cx(b[0], b[1])
        >>> p(0.5, b[1])
        >>> print(a.qs)
        QuantumCircuit:
        ---------------
                  ┌───┐               
        a.0: ──■──┤ H ├───────────────
               │  ├───┤               
        a.1: ──■──┤ H ├───────────────
               │  └───┘               
        a.2: ──■──────────────────────
             ┌─┴─┐┌───┐               
        b.0: ┤ X ├┤ X ├──■────────────
             └───┘└───┘┌─┴─┐┌────────┐
        b.1: ──────────┤ X ├┤ P(0.5) ├
                       └───┘└────────┘
        Live QuantumVariables:
        ----------------------
        QuantumVariable a
        QuantumVariable b
        
        >>> b.uncompute()
        >>> print(b.qs)
        QuantumCircuit:
        ---------------
             ┌────────┐                              ┌────────┐┌───┐
        a.0: ┤0       ├──────────────────────────────┤0       ├┤ H ├
             │        │                              │        │├───┤
        a.1: ┤1       ├──────────────────────────────┤1       ├┤ H ├
             │  pt3cx │                              │  pt3cx │└───┘
        a.2: ┤2       ├──────────────────────────────┤2       ├─────
             │        │┌───┐                    ┌───┐│        │     
        b.0: ┤3       ├┤ X ├──■──────────────■──┤ X ├┤3       ├─────
             └────────┘└───┘┌─┴─┐┌────────┐┌─┴─┐└───┘└────────┘     
        b.1: ───────────────┤ X ├┤ P(0.5) ├┤ X ├────────────────────
                            └───┘└────────┘└───┘                    
        Live QuantumVariables:
        ----------------------
        QuantumVariable a


        """
        
        if self.is_deleted():
            raise Exception("Tried to uncompute deleted QuantumVariable")
        
        
        if do_it:
            from qrisp.uncomputation import uncompute
            uncompute(self.qs, self.qs.uncomp_stack + [self], recompute)
            self.qs.uncomp_stack = []
        else:
            self.qs.uncomp_stack.append(self)
        
        
        
    def get_unique_name(self, name = None):
        
        if name is None:
            from qrisp import QuantumBool, QuantumFloat, QuantumChar    
            if isinstance(self, QuantumBool):
                name = "qbl"
            elif isinstance(self, QuantumFloat):
                name = "qf"
            elif isinstance(self, QuantumChar):
                name = "qch"
            else:
                name = "qv"
        
        while True:
            try:
                naming_number = self.name_tracker[name]
                self.name_tracker[name] += 1
                name = name + "_" + str(naming_number)
            except KeyError:
                self.name_tracker[name] = 1
                name = name + "_0"
            
            i = 0
            while i < len(QuantumVariable.live_qvs):
                qv = QuantumVariable.live_qvs[i]()
                if qv is None:
                    QuantumVariable.live_qvs.pop(i)
                    continue
                if qv.name == name:
                    break
                i += 1
            else:
                break
            
        return name
        
    def init_from(self, other):
        """
        Method to initiate a QuantumVariable based on the state of another. This method does NOT copy the state. Much rather it performs the operation
        
        
        .. math::
            
            U_{\\text{init_from}} \\left( \sum_{x \\in \\text{labels}} a_x \ket{x} \\right)  \ket{0} = \sum_{x \\in \\text{labels}} a_x \ket{x} \ket{x}
        
        
        This is different from a state copying operation: 
        
        .. math::
            
            U_{\\text{copy}} \\left( \sum_{x \\in \\text{labels}} a_x \ket{x} \\right)  \ket{0} = \\left( \\sum_{x \\in \\text{labels}} a_x \ket{x} \\right) \\left( \\sum_{x \\in \\text{labels}} a_x \ket{x} \\right)
            
        
        A shorthand for initiating this way is the ``[:]`` operator.
        

        Parameters
        ----------
        other : QuantumVariable
            The QuantumVariable from which to initiate.

        Raises
        ------
        Exception
            Tried to initialize qubits which are not fresh anymore.
        
        Examples
        --------
        
        We create a QuantumFloat, and bring it into superposition. 
        
        >>> from qrisp import QuantumFloat, h, multi_measurement
        >>> qf_a = QuantumFloat(8)
        >>> qf_a[:] = 6
        >>> h(qf_a[0])
        >>> print(qf_a)
        {6: 0.5, 7: 0.5}
        
        We now duplicate and initiate the duplicate
        
        >>> qf_b = qf_a.duplicate()
        >>> print(qf_b)
        {0: 1.0}
        >>> qf_b.init_from(qf_a)
        >>> print(multi_measurement([qf_a, qf_b]))
        {(6, 6): 0.5, (7, 7): 0.5}
        
        The slicing operator achieves the same:
            
        >>> qf_c = qf_a.duplicate()
        >>> qf_c[:] = qf_b
        >>> print(multi_measurement([qf_a, qf_b, qf_c]))
        {(6, 6, 6): 0.5, (7, 7, 7): 0.5}
        
        
        
        """
        
        if not type(self) == type(other):
            raise Exception("Tried to initialize " + str(type(self)) + " from " + str(type(other)))
        
        from qrisp.misc import check_if_fresh
        
        if not check_if_fresh(self.reg, self.qs):
            raise Exception("Tried to initialize qubits which are not fresh anymore")
        
        self.qs.cx(other.reg, self.reg)
    
    @classmethod
    def custom(self, label_list, decoder = None, qs = None, name = None):
        """
        Creates a QuantumVariable with customized outcome labels.
        
        Note that this is a class method, implying there is no need to create another QuantumVariable first to call this method.

        Parameters
        ----------
        label_list : list
            A list of outcome labels.
        decoder : function, optional
            The decoder function. If given none, the labels will be encoded according to their placement in the ``label_list``.
        qs : QuantumSession, optional
            The :ref:`QuantumSession` in which to register the customized QuantumVariable. If given none, the QuantumVariable will be registered in a new QuantumSession.
        name : string, optional
            The name of the QuantumVariable. If given none, a suited name will be generated.

        Returns
        -------
        CustomQuantumVariable
            A QuantumVariable with the desired outcome labels.

        Examples
        --------
        
        We create a QuantumVariable with some examples values as outcome labels and bring it into uniform superposition.
        
        >>> from qrisp import QuantumVariable, h
        >>> qv = QuantumVariable.custom(["lorem", "ipsum", "dolor", "sit", 42, (1,2,3)])
        >>> h(qv)
        >>> print(qv)
        {'lorem': 0.125, 'ipsum': 0.125, 'dolor': 0.125, 'sit': 0.125, 42: 0.125, (1, 2, 3): 0.125, 'undefined_label_6': 0.125, 'undefined_label_7': 0.125}

        """
        from qrisp.misc import custom_qv
        return custom_qv(label_list, decoder = decoder, qs = qs, name = name)
        


def plot_histogram(outcome_labels, counts, filename = None):
    
    res_list = []
    
    for k in range(len(outcome_labels)):
        try:
            res_list.append(counts[outcome_labels[k]])
        except KeyError:
            res_list.append(0)
    
    plt.bar(outcome_labels, res_list, width = 1/len(outcome_labels))
    plt.grid()
    plt.ylabel("Measurement probability")
    plt.xlabel("QuantumVariable value")
    
    if filename:
        plt.savefig(filename, dpi = 400, bbox_inches = "tight")
    else:
        plt.show()



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# -*- coding: utf-8 -*-

#This file contains some functions that enhance the speed
#of SparseBiArray manipulation

import numpy as np
from numba import njit, int64, vectorize, prange

#It can happen that the coo-matrix multiplication puts two data entries
#for the same index. These function recognizes and sums these duplicates
def sum_duplicates(nz_indices, data):
    
    new_indices, reconstr_indices = np.unique(nz_indices, return_inverse = True)
    
    if len(nz_indices) > 1E4:
        new_data = sum_duplicates_aux_jitted(new_indices, data, reconstr_indices)
    else:
        new_data = sum_duplicates_aux(new_indices, data, reconstr_indices)
    
    return new_indices, new_data


def sum_duplicates_aux(new_indices, old_data, reconstr_indices):
    
    new_data = np.zeros(new_indices.shape, dtype = old_data.dtype)
    
    for i in range(len(reconstr_indices)):
        new_data[reconstr_indices[i]] += old_data[i]
    
    return new_data



@njit(nogil = True, cache = True)
def sum_duplicates_aux_jitted(new_indices, old_data, reconstr_indices):
    
    new_data = np.zeros(new_indices.shape, dtype = old_data.dtype)
    
    for i in range(len(reconstr_indices)):
        new_data[reconstr_indices[i]] += old_data[i]
    
    return new_data





#This function rounds entries, which are below a certain float thresh and removes the tracked indices
# @njit(parallel = True)
def elim_zeros(nz_indices, data, decimals = 8):
    
    # return nz_indices, data
    data = np.around(data, decimals = decimals, out = data)
    
    # mask = np.abs(data) > 10**-decimals
    mask = np.nonzero(data)[0]
    
    return nz_indices[mask], data[mask]



#This function applies an index permutation onto
#flattened indices.
#Consider the situation that we have the two flattened indices
#[1,3]
#of a 2x2 matrix.
#If we now want to permute the axes, we permute the bit level
#representation of these numbers
#["01", "10"] ==> ["10", "01"]

@njit(nogil = True, cache = True)
def permute_axes(index, permutation):
    
    permutation = -permutation + len(permutation) -1
    permutation = permutation[::-1]
    new_index = np.zeros(index.size, dtype = np.int64)
    for i in range(len(permutation)):
        
        if permutation[i] == i:
            # new_index = new_index | (index&(1<<permutation[i]))
            for j in prange(len(new_index)):
                new_index[j] = new_index[j] | (index[j]&(1<<permutation[i]))
        else:
            # new_index = new_index | (((index&(1<<permutation[i]))>>permutation[i])<<(i))
            for j in prange(len(new_index)):
                new_index[j] = new_index[j] | (((index[j]&(1<<permutation[i]))>>permutation[i])<<(i))
    
    return new_index


#This function retrieves the coordinates from
#flattened coordinades for a given shape of length two
def get_coordinates(indices, shape):
    return get_coordinates_0(indices, int(np.log2(shape[1]))), get_coordinates_1(indices, int(np.log2(shape[1])))


@vectorize([int64(int64, int64)])
def get_coordinates_0(index, bit_shape_1):
    return index>>bit_shape_1


@vectorize([int64(int64, int64)])
def get_coordinates_1(index, bit_shape_0):
    return index&((1<<bit_shape_0)-1)

#This function basically inverts get_coordinates.
#Ie. it generates flat coordinates from row and column indices
def gen_flat_coords(col, row, shape):
    return gen_flat_coords_vec(col, row, int(np.log2(shape[1])))


@vectorize([int64(int64, int64, int64)], cache = True)
def gen_flat_coords_vec(col, row, bit_shape_1):
    return row<<bit_shape_1 | col

#This function constructs a flattened numpy array from
#given flattened coordinates and data
@njit(nogil = True, cache = True)
def construct_flat_array(flat_coords, data, size):
    
    res = np.zeros(size, dtype = data.dtype)
    
    for i in range(len(flat_coords)):
        res[flat_coords[i]] += data[i]
    
    return res



@njit(cache = True)
def extract_bit_range(index, start, stop):
    breadth = stop - start
    return ((index&(((1<<breadth)-1)<<start))>>start)


@njit(cache = True)
def insert_bit_range(target, bit_string, start):
    target |= (bit_string<<start)

@njit(nogil = True, cache = True)
def jitted_permuter(index, new_index, bit_partition, permuted_bit_partition, perm):
    
    # for i in range(len(perm)):
    #     # bit_range = extract_bit_range(index, bit_partition[i], bit_partition[i+1])
    #     # insert_bit_range(new_index, bit_range, permuted_bit_partition[perm[i]])
    #     #Inlined version more efficient:
    #     new_index |= (((index&((((1<<(bit_partition[i+1] - bit_partition[i]))-1))<<bit_partition[i]))>>bit_partition[i])<<permuted_bit_partition[perm[i]])
    
    for j in prange(index.size):
        for i in range(len(perm)):
            # bit_range = extract_bit_range(index, bit_partition[i], bit_partition[i+1])
            # insert_bit_range(new_index, bit_range, permuted_bit_partition[perm[i]])
            #Inlined version more efficient:
            new_index[j] |= (((index[j]&((((1<<(bit_partition[i+1] - bit_partition[i]))-1))<<bit_partition[i]))>>bit_partition[i])<<permuted_bit_partition[perm[i]])
            
        

def permuter(index, new_index, bit_partition, permuted_bit_partition, perm):
    for i in range(len(perm)):
        # bit_range = extract_bit_range(index, bit_partition[i], bit_partition[i+1])
        # insert_bit_range(new_index, bit_range, permuted_bit_partition[perm[i]])
        #Inlined version more efficient:
        new_index |= (((index&((((1<<(bit_partition[i+1] - bit_partition[i]))-1))<<bit_partition[i]))>>bit_partition[i])<<permuted_bit_partition[perm[i]])



        
        
def invert_perm(perm):
    return [perm.index(i) for i in range(len(perm))]


def permute_axes(index, index_bit_permutation):
    
    n = len(index_bit_permutation)


    index_bit_permutation = -np.array(index_bit_permutation) + n -1
    index_bit_permutation = list(index_bit_permutation)[::-1]    
    
    
    index_bit_permutation = list(index_bit_permutation)

    
    #Get the permutation of the axes
    
    perm = [index_bit_permutation.index(i) for i in range(n)]



    #Initialize the shape
    perm_shape = n*[1]
    
    i = 1
    while i < len(perm):

        
        #If this condition is met, both indices form an interval,
        #that does not need further permutation
        if perm[i]-1 == perm[i-1]:
            #Remove the index from the perm and perm_shape list
            perm.pop(i)
            perm_shape.pop(i)
            
            #Update the shape
            perm_shape[i-1] += 1
            #Update the permutation indices
            for j in range(len(perm)):
                if perm[j] > perm[i-1]:
                    perm[j] -= 1
        else:
            i += 1
    
    
    #print(perm)
    new_index = np.zeros(index.size, dtype = np.int64)

    bit_partition = [sum(perm_shape[:i]) for i in range(len(perm_shape)+1)]

    #print(bit_partition)
    perm_shape_permuted = [perm_shape[i] for i in invert_perm(perm)]
    permuted_bit_partition = [sum(perm_shape_permuted[:i]) for i in range(len(perm_shape)+1)]
    
    if len(perm)*index.size > 2**10:
        jitted_permuter(index, new_index, np.array(bit_partition), np.array(permuted_bit_partition), np.array(perm))
    else:
        permuter(index, new_index, np.array(bit_partition), np.array(permuted_bit_partition), np.array(perm))
    
    return new_index

@njit(cache = True)
def invert_permutation(p):
    """Return an array s with which np.array_equal(arr[p][s], arr) is True.
    The array_like argument p must be some permutation of 0, 1, ..., len(p)-1.
    """
    s = np.empty(len(p), dtype = np.int64)
    s[p] = np.arange(p.size)
    return s


@njit(nogil = True, cache = True)
def bi_array_moveaxis(data_array, index_perm, f_index_array):
    
    for i in range(len(index_perm)):
        shifted_value = 1<<index_perm[i]
        start_index = 1<<i
        
        for j in range(start_index):
            f_index_array[start_index ^ j] = shifted_value ^ f_index_array[j]
    
    # return f_index_array
    return data_array[f_index_array]

        


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# Created by ann81984 at 26.07.2022
from qrisp import QuantumVariable, QuantumFloat
import numpy as np
import matplotlib.pyplot as plt
from numpy.linalg import norm

def test_state_preparation():
    qf = QuantumFloat(4, -2, signed = True)


    state_dic = {2.75 : 1/4**0.5,
                -1.5  : -1/4**0.5,
                2     : 1/4**0.5,
                3     : 1j/4**0.5}


    qf.init_state(state_dic)

    debugger = qf.qs.statevector("function")
    
    print(type(debugger))

    print("Amplitude of state 2.75: ", debugger({qf : 2.75}))
    print("Amplitude of state -1.5: ", debugger({qf : -1.5}))
    print("Amplitude of state 3: ", debugger({qf : 3}))

    assert np.abs(debugger({qf: 2.75}) - 0.5) < 1E-5
    assert np.abs(debugger({qf: -1.5}) + 0.5) < 1E-5
    assert np.abs(debugger({qf: 3}) - 0.5j) < 1E-5


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


from qrisp.core.quantum_variable import QuantumVariable
from qrisp.environments import control
import sys
class QuantumBool(QuantumVariable):
    """
    QuantumBools are the quantum type, which represents boolean truth values.
    They are the return type of comparison operators like the equality ``==``.
    
    Apart from their behavior as a QuantumVariable, they can also be treated like :ref:`ControlEnvironments <ControlEnvironment>`.
    
    .. note:
        QuantumBools that are evaluated directly after a ``with`` statement are uncomputed automatically upon leaving the ControlEnvironment.
        
    Examples
    --------
    
    We create a QuantumBool and set it to uniform superposition
    
    >>> from qrisp import QuantumBool, h
    >>> q_bool_0 = QuantumBool()
    >>> h(q_bool_0)
    >>> print(q_bool_0)
    {False: 0.5, True: 0.5}
    
    We create a second QuantumBool and evaluate some logical functions
    
    >>> q_bool_1 = QuantumBool()
    >>> print(q_bool_1 | q_bool_0)
    {False: 0.5, True: 0.5}
    >>> print(q_bool_1 & q_bool_0)
    {False: 1.0}
    
    QuantumBools are the results of comparisons:
    
    >>> from qrisp import QuantumFloat, QuantumChar
    >>> q_ch = QuantumChar()
    >>> q_ch[:] = {"g" : 1, "l" : -1}
    >>> q_bool_2 = (q_ch == "g")
    >>> q_bool_2.qs.statevector()
    sqrt(2)*(|g>*|True> - |l>*|False>)/2    
    
    For :ref:`QuantumFloats <QuantumFloat>`, numeric comparison is also possible:
    
    
    >>> qf = QuantumFloat(4)
    >>> h(qf[3])
    >>> print(qf)
    {0: 0.5, 8: 0.5}
    >>> q_bool_3 = (qf >=  4)
    >>> print(q_bool_3)
    {False: 0.5, True: 0.5}
    
    To use a QuantumBool as a :ref:`ControlEnvironment`, we simply put it in a ``with`` statement:
        
    ::
        
        with q_bool_3:
            qf += 2
    
    >>> print(qf)
    {0: 0.5, 10: 0.5}
    
    QuantumBools that are created directly after a ``with`` statement are uncomputed automatically:
    
    ::
        
        with qf == 10:
            q_bool_3.flip()
            
    >>> print(qf.qs)
    QuantumCircuit:
    ---------------
                       ┌────────────┐     ┌───────────┐               
            qf.0: ─────┤0           ├─────┤0          ├──o─────────o──
                       │            │     │           │  │         │  
            qf.1: ─────┤1           ├─────┤1          ├──■─────────■──
                       │            │     │  __iadd__ │  │         │  
            qf.2: ─────┤2           ├─────┤2          ├──o─────────o──
                  ┌───┐│  less_than │     │           │  │         │  
            qf.3: ┤ H ├┤3           ├─────┤3          ├──■─────────■──
                  └───┘│            │┌───┐└─────┬─────┘  │  ┌───┐  │  
        lt_qbl.0: ─────┤4           ├┤ X ├──────■────────┼──┤ X ├──┼──
                       │            │└───┘               │  └─┬─┘  │  
    lt_ancilla.0: ─────┤5           ├────────────────────┼────┼────┼──
                       └────────────┘                  ┌─┴─┐  │  ┌─┴─┐
      cond_env.0: ─────────────────────────────────────┤ X ├──■──┤ X ├
                                                       └───┘     └───┘
    Live QuantumVariables:
    ----------------------
    QuantumFloat qf
    QuantumBool lt_qbl
        
    Note that there is only a single QuantumBool listed in the "Live QuantumVariables" section, 
    because the QuantumBool of the comparison ``qf == 10`` (called ``cond_env``) has been uncomputed.

    """
    

    def __init__(self, qs = None, name = None):

        QuantumVariable.__init__(self, 1, qs = qs, name = name)
        
        self.qfloat_comparison = False
        
    def decoder(self, integer):
        return bool(integer)
    
    def __and__(self, other):
        from qrisp import mcx
        
        and_qbl = QuantumBool()
        
        mcx(self.reg + other.reg, and_qbl.reg)
        
        return and_qbl

    def __or__(self, other):
        from qrisp import mcx, x
        
        or_qbl = QuantumBool()

        x(self)
        x(other)
        
        mcx(self.reg + other.reg, or_qbl.reg)
        
        x(self)
        x(other)
        
        x(or_qbl)
        
        return or_qbl
    
    def __xor__(self, other):
        
        from qrisp import cx
        
        xor_qbl = QuantumBool()
        
        cx(self, xor_qbl)
        cx(other, xor_qbl)
        
        return xor_qbl
    
    def flip(self):
        """
        Flips the QuantumBool's value.

        """
        
        from qrisp import x
        
        x(self)
        return self
    
    def __enter__(self):
        
        self.env = control(self)
        self.env.__enter__()

    def __exit__(self, a, b, c):
        
        ref_count = sys.getrefcount(self)
        
        #If the refcount is 5, this means that the QuantumBool has been
        #created in a "with" statement and can therefore no longer be reached
        #after exiting. We therefore uncompute
        
        self.env.__exit__(a, b, c)
        if ref_count == 5:
            self.uncompute()



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""


#Modification to Kahns algorithm to reduce the depth of general circuits by applying
#trivial and non-trivial commutation relations based on the dag representation of unqomp

import networkx as nx
import numpy as np
from numba import njit

from qrisp.uncomputation.unqomp import dag_from_qc

def qb_set_to_int(qubits, qc):
    res = 0
    for qb in qubits:
        res |= 1<<qc.qubits.index(qb)
    return res


#Kahns Algorithm based on https://www.geeksforgeeks.org/topological-sorting-indegree-based-solution/

def depth_sensitive_topological_sort(indices, indptr, int_qc, num_qubits):
      
    # Create a vector to store indegrees of all
    # vertices. Initialize all indegrees as 0.
    n = len(indptr)-1
    in_degree = np.zeros(n, dtype = np.int32)
    
    depths = np.zeros(num_qubits, dtype = np.int32)
    # Traverse adjacency lists to fill indegrees of
    # vertices.  This step takes O(V + E) time

     
    for i in range(n):
         for j in indices[indptr[i]:indptr[i+1]]:
             in_degree[j] += 1
 
    # Create a queue and enqueue all vertices with
    # indegree 0
    queue = []

    for i in range(n):
        if in_degree[i] == 0:
            queue.append(i)
 
    
    # Initialize count of visited vertices
    cnt = 0
 
    # Create a vector to store result (A topological
    # ordering of the vertices)
    top_order = np.zeros(n, dtype = np.int32)
 
    # One by one dequeue vertices from queue and enqueue
    # adjacents if indegree of adjacent becomes 0
     
    while queue:
 
         #The depth sensitive part is now to deque the node with the least depth
        node_costs = np.zeros(len(queue))
        
        for i in range(len(queue)):
            node = queue[i]
            
            qubits = int_qc[node]
            
            depth_sum = 0
            for j in range(num_qubits):
                if qubits & 1<<j:
                    depth_sum += depths[j]
            
            node_costs[i] = depth_sum

        u = queue.pop(np.argmin(node_costs))
        
        top_order[cnt] = u
 
        #Update depths array
        max_depth = 0
        
        for i in range(num_qubits):
            if int_qc[u] & 1<<i:
                if depths[i] > max_depth:
                    max_depth = depths[i]
        
        for i in range(num_qubits):
            if int_qc[u] & 1<<i:
                depths[i] = max_depth + 1

    
        #Update in degree array
        for i in indices[indptr[u]:indptr[u+1]]:
            in_degree[i] -= 1
            if in_degree[i] == 0:
                queue.append(i)
        
        cnt += 1
         
    return top_order
        

depth_sensitive_topological_sort_jitted = njit(cache = True)(depth_sensitive_topological_sort)

def reduce_depth(qc):

    if len(qc.data) <= 1:
        return qc.copy()
    
    dag = dag_from_qc(qc, remove_init_nodes = True)
    
    sprs_mat = nx.to_scipy_sparse_array(dag, format = "csr")
    
    node_list = list(dag.nodes())
    qubit_ints = [qb_set_to_int(n.instr.qubits, qc) for n in node_list]
    
    try:
        qubit_ints = np.array(qubit_ints, dtype = np.int64)
        res = depth_sensitive_topological_sort_jitted(sprs_mat.indices, sprs_mat.indptr, qubit_ints, num_qubits = qc.num_qubits())
    except OverflowError:
        res = depth_sensitive_topological_sort(sprs_mat.indices, sprs_mat.indptr, qubit_ints, num_qubits = qc.num_qubits())
    
    qc_new = qc.clearcopy()
    
    for i in range(len(res)):
        qc_new.append(node_list[res[i]].instr)
        
    return qc_new



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""



import numpy as np
from qrisp.circuit.operation import PTControlledOperation, ControlledOperation
from qrisp.circuit.instruction import Instruction
from qrisp.circuit.operation import Operation
from qrisp.circuit.quantum_circuit import fast_append

#This function dissolves any Operation objects that have a definition circuit such that the result only consists of elementary gates
def transpile(qc, transpilation_level = np.inf, transpile_predicate = None, **kwargs):
    
    from qrisp.circuit import Operation, QuantumCircuit, Qubit, Clbit

    with fast_append():
        from qrisp.circuit import Qubit, Clbit
        transpiled_qc = QuantumCircuit()
        
        # [transpiled_qc.add_qubit(Qubit(qb.identifier)) for qb in qc.qubits]
        # [transpiled_qc.add_clbit(Clbit(cb.identifier)) for cb in qc.clbits]
        
        for qb in qc.qubits:
            if isinstance(qb, Qubit):
                transpiled_qc.add_qubit(qb)
            else:
                transpiled_qc.add_qubit(Qubit(qb.identifier))
                
        for cb in qc.clbits:
            if isinstance(qb, Qubit):
                transpiled_qc.add_clbit(cb)
            else:
                transpiled_qc.add_clbit(Clbit(cb.identifier))
    
        translation_dic = {qc.qubits[i].identifier : transpiled_qc.qubits[i] for i in range(len(qc.qubits))}
        translation_dic.update({qc.clbits[i].identifier : transpiled_qc.clbits[i] for i in range(len(qc.clbits))})
        
        if transpile_predicate is None:
            transpile_predicate_ = lambda i, op : i < transpilation_level
        else:
            transpile_predicate_ = lambda i, op : i < transpilation_level and transpile_predicate(op)
        
        transpile_inner(qc, transpiled_qc, translation_dic, transpile_predicate_)
        
        QuantumCircuit.fast_append = False
        
        if not kwargs or not hasattr(qc, "to_qiskit"):
            return transpiled_qc
        else:
            from qrisp import QuantumCircuit
            qiskit_qc = transpiled_qc.to_qiskit()
            
            from qiskit import transpile as qiskit_transpile
            
            transpiled_qiskit_qc = qiskit_transpile(qiskit_qc, **kwargs)
            
            qrisp_qc = QuantumCircuit.from_qiskit(transpiled_qiskit_qc)
            
            return qrisp_qc


def transpile_inner(transpilation_qc, target_qc, translation_dic, transpile_predicate, transpilation_level = 0):
    
    for i in range(len(transpilation_qc.data)):
        instr = transpilation_qc.data[i]
        if instr.op.definition:
            
            if transpile_predicate(transpilation_level, instr.op):
                
                definition = instr.op.definition
                
                new_translation_dic = {definition.qubits[j].identifier : translation_dic[instr.qubits[j].identifier] for j in range(len(instr.qubits))}
                
                new_translation_dic.update({definition.clbits[j].identifier : translation_dic[instr.clbits[j].identifier] for j in range(len(instr.clbits))})
                
                transpile_inner(instr.op.definition, 
                                target_qc, 
                                new_translation_dic, 
                                transpile_predicate, 
                                transpilation_level + 1)
                continue
        
        if not isinstance(instr.op, Operation):
            op = Operation(init_op = instr.op)
        else:
            op = instr.op
        
        target_qc.data.append(Instruction(op, [translation_dic[qb.identifier] for qb in instr.qubits], [translation_dic[cb.identifier] for cb in instr.clbits]))
        # target_qc.append()
        
    


def extend(qc_0, qc_1, translation_dic = "id"):
    
    if translation_dic == "id":
        
        translation_dic = {}
        for qb in qc_1.qubits:
            translation_dic[qb] = qb
            if not qb in qc_0.qubits:
                qc_0.add_qubit(qb)
            
        for cb in qc_1.clbits:
            translation_dic[qb] = qb
            if not cb in qc_0.clbits:
                qc_0.add_clbit(cb)
            

    #Copy in order to prevent modification
    translation_dic = dict(translation_dic)

    for key in list(translation_dic.keys()):
        
        if not isinstance(key, str):
            translation_dic[key.identifier] = translation_dic[key]


    for i in range(len(qc_1.data)):
        
        instruction_other = qc_1.data[i]
        qubits = []
        for qb in instruction_other.qubits:
            qubits.append(translation_dic[qb.identifier])
        
        clbits = []
        
        for cb in instruction_other.clbits:
            clbits.append(translation_dic[cb.identifier])
        
        instr_type = type(instruction_other)
        
        qc_0.data.append(instr_type(instruction_other.op, qubits, clbits))
        


"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# -*- coding: utf-8 -*-

import threading
from qrisp.simulator.quantum_state import QuantumState

import numpy as np
thread_limit = 1

#This class describes a set of decoherent quantum states
class ImpureQuantumState:
    
    
    #Initially we start we a single quantum state
    def __init__(self, n, clbit_amount = 0):
        self.states = [QuantumState(n)]
        self.cl_prob = [1]
        self.outcome_list = [np.zeros(clbit_amount, dtype = np.byte)]
        self.collected_measurements = []
        
    
    
    def apply_operation(self, operation, qubits):
        
        #Perform operation application in parallel. This is possible because many
        #functions in here release the GIL (such as numpy's np.dot)
        if thread_limit > 1 and len(self.states) > 1:
            
            threads = []
            
            for i in range(len(self.states)):
                mp_wrapper = lambda : self.states[i].apply_operation(operation, qubits)
                
                thr = threading.Thread(target = mp_wrapper)
                
                if len(threads) >= thread_limit:
                    threads[0].join()
                    threads.pop(0)
                
                thr.start()
                threads.append(thr)
                
            
            [thr.join() for thr in threads]
            
        else:
            
            for i in range(len(self.states)):
                self.states[i].apply_operation(operation, qubits)
            

    def conditionally_apply_operation(self, operation, qubits, clbit):
        
        #Perform operation application in parallel. This is possible because many
        #functions in here release the GIL (such as numpy's np.dot)
        
        if thread_limit > 1:
            
            threads = []
            
            for i in range(len(self.states)):
                
                if self.outcome_list[clbit] == "0":
                    continue
                
                mp_wrapper = lambda : self.states[i].apply_operation(operation, qubits)
                
                if len(threads) >= thread_limit:
                    threads[0].join()
                    threads.pop(0)
                
                
                thr = threading.Thread(target = mp_wrapper)
                thr.start()
                
                threads.append(thr)
            
            [thr.join() for thr in threads]
            
        else:
            
            for i in range(len(self.states)):
                if self.outcome_list[clbit] == "0":
                    continue

                
                self.states[i].apply_operation(operation, qubits)

        
    #Perform reset operation
    #We perform the reset operation by measuring but not tracking the measurement result
    def reset(self, qubit_index, keep_outcome = False):
        
        if thread_limit > 1:
            
            threads = []
            
            for i in range(len(self.states)):
                mp_wrapper = lambda : self.states[i].measure(qubit_index, keep_outcome)
                thr = threading.Thread(target = mp_wrapper)
                
                if len(threads) > thread_limit:
                    threads[0].join()
                    threads.pop(0)
                
                
                thr.start()
                threads.append(thr)
                
            
        else:
            
            for i in range(len(self.states)):
                self.states[i].measure(qubit_index, keep_outcome)
        
        
        #Generate new list of states and the corresponding classical propabilities + the outcome bitstrings
        new_states = []
        new_cl_prob = []
        new_outcome_list = []
        
        for i in range(len(self.states)):
            
            #Wait for measurement threads to finish
            if thread_limit > 1:
                if len(threads):
                    threads[0].join()
                    threads.pop(0)

            #Get measurement outcome
            #p_0 represents the probability to measure a 0,
            #state_0 represents the state after measuring a 0
            p_0, state_0, p_1, state_1 = self.states[i].last_mes_outcome
            #Log results
            if p_0 != 0:
                
                #Add the new decoherent state to the list of states
                new_states.append(state_0)
                
                #Update its classical probability
                new_cl_prob.append(p_0*self.cl_prob[i])
                
                #The bitstring stays the same
                new_outcome = self.outcome_list[i].copy()
                new_outcome_list.append(new_outcome)
                
            #Perform the same thing for the outcome 1
            if p_1 != 0:
                new_states.append(state_1)
                new_cl_prob.append(p_1*self.cl_prob[i])
                new_outcome = self.outcome_list[i].copy()
                new_outcome_list.append(new_outcome)
        
        
        #Update attributes
        self.states = new_states
        self.cl_prob = new_cl_prob
        self.outcome_list = new_outcome_list
        
        
    #This functions performs measurements
    #The logic behind it is rather similar to reset, but this time we log the measurement outcome
    def measure(self, qubit_index, clbit_index):
        
        if thread_limit > 1:
            
            threads = []
            
            for i in range(len(self.states)):
                mp_wrapper = lambda : self.states[i].measure(qubit_index)
                thr = threading.Thread(target = mp_wrapper)
                
                if len(threads) > thread_limit:
                    threads[0].join()
                    threads.pop(0)
                
                
                thr.start()
                threads.append(thr)
            
        else:
            
            for i in range(len(self.states)):
                self.states[i].measure(qubit_index, True)

        
        new_states = []
        new_cl_prob = []
        new_outcome_list = []
        
        for i in range(len(self.states)):
            
            #Wait for measurement threads to finish
            if thread_limit > 1:
                if len(threads):
                    threads[0].join()
                    threads.pop(0)
            
            p_0, state_0, p_1, state_1 = self.states[i].last_mes_outcome
            
            if p_0 != 0:
                new_states.append(state_0)
                new_cl_prob.append(p_0*self.cl_prob[i])
                new_outcome = self.outcome_list[i].copy()
                
                #Here is the point, where we log the measurement outcome
                new_outcome[clbit_index] = 0
                new_outcome_list.append(new_outcome)
            if p_1 != 0:
                
                
                new_states.append(state_1)
                new_cl_prob.append(p_1*self.cl_prob[i])
                new_outcome = self.outcome_list[i].copy()
                new_outcome[clbit_index] = 1
                new_outcome_list.append(new_outcome)
                
        self.states = new_states
        self.cl_prob = new_cl_prob
        self.outcome_list = new_outcome_list
    
    
    def add_qubit(self):
        for state in self.states: state.add_qubit()
        
    def add_clbit(self):
        
        for i in range(len(self.outcome_list)):
            self.outcome_list[i] = np.array([0] + list(self.outcome_list[i]), dtype = np.byte)



"""
/*********************************************************************
* Copyright (c) 2023 the Qrisp Authors
*
* This program and the accompanying materials are made
* available under the terms of the Eclipse Public License 2.0
* which is available at https://www.eclipse.org/legal/epl-2.0/
*
* SPDX-License-Identifier: EPL-2.0
**********************************************************************/
"""

# -*- coding: utf-8 -*-

from qrisp.simulator.numerics_config import xp, float_tresh
from qrisp.simulator.bi_arrays import SparseBiArray, DenseBiArray, tensordot

#tensordot = xp.tensordot

#Class to describe TensorFactors

class TensorFactor:
    
    def __init__(self, qubits, init_tensor_array = None):
        
        #This list contrains a list of integers, which describe the current permutation of qubit indices inside this TensorFactor
        #ie. swapping two elements of this list is the same as applying a swap gate. Using this structure is advantageous
        #because applying matrices on specific qubit indices requires alot of swapping
        #Swapping the indices of an array is ressource costly because the array needs to be copied.
        #As it turns out memory speed is a bottleneck.
        self.qubits = qubits
        
        self.n = len(qubits)
        
        if init_tensor_array is None:
            self.tensor_array = xp.zeros(2**self.n, dtype = xp.complex64)
            self.tensor_array[0] = 1
            
        else:
            self.tensor_array = init_tensor_array
        
        if isinstance(self.tensor_array, xp.ndarray):
            self.tensor_array = SparseBiArray(self.tensor_array)
        
       
    def copy(self):
        return TensorFactor(list(self.qubits), self.tensor_array.copy())
    
    #Swaps two qubit indices
    def swap(self, i, j):
        
        if i == j:
            return
        
        self.qubits[i], self.qubits[j] = self.qubits[j], self.qubits[i]
        
        self.tensor_array = self.tensor_array.reshape(self.n*[2])
        self.tensor_array = self.tensor_array.swapaxes(i, j)
        self.tensor_array = self.tensor_array.reshape(2**self.n)
    
    
    #Applies a matrix onto the given state
    def apply_matrix(self, matrix, qubits):
        
        
        #Remove the non-zero entries caused by floating point errors to get 
        #the most out of the sparse matrix representation
        import numpy as np
        if matrix.size >= 2**6:
            
            if matrix.dtype == np.complex64:
                matrix = matrix*(np.abs(matrix) > 1E-7)                
            if matrix.dtype == np.complex128:
                matrix = matrix*(np.abs(matrix) > 1E-15) 
                
        #Convert matrix to BiArray
        matrix = DenseBiArray(matrix)
        
        #Swap the qubits which the matrix is supposed to be applied on to the front
        for i in range(len(qubits)):
            qubit_index = self.qubits.index(qubits[i])
            self.swap(qubit_index, i)
        
        #Reshape tensor array such that the qubits form a single index
        reshaped_tensor_array = self.tensor_array.reshape((matrix.shape[0], 2**self.n//matrix.shape[0]))
        
        #Apply matrix
        self.tensor_array = tensordot(matrix, reshaped_tensor_array,  axes = (1,0), contract_sparsity_threshold = 0.05).reshape(2**self.n)
        # print(type(self.tensor_array))
    
    #Entangles two TensorFactors.
    def entangle(self, other):
        
        if self == other:
            return self
        if set(self.qubits).intersection(other.qubits):
            raise Exception("Tried to entangle two tensor factors with overlapping qubits")
        
        #The tensordot function with axes = ((),()) performs a,b -> a (x) b, where (x) is the tensor product sign
        init_array = tensordot(other.tensor_array, self.tensor_array, axes = ((), ()), contract_sparsity_threshold = 0.01).ravel()
        
        return TensorFactor(other.qubits + self.qubits, init_array)


    #Perform measurement
    def measure(self, i):
        
        #Swap the index that is supposed to be measured to the front
        index = self.qubits.index(i)
        self.swap(0, index)
        
        #Split the array - the lower half corresponds to the state with outcome 0, the upper half to 1        
        lower_half, upper_half = self.tensor_array.split()        
        
        #Determine probabilities
        p_0 = lower_half.squared_norm()
        p_1 = upper_half.squared_norm()
        
        #p_1 = abs(p_0 - 1) #This doesn't work due to floating point errors accumulating in the normalization
        
        #Prepare the qubit list of the resulting TensorFactor
        #After the measurement, the measured qubit is disentangled from the rest,
        #so we can put it into it's own TensorFactor object
        new_qubit_list = list(self.qubits)
        new_qubit_list.pop(0)
        
        #This treats the case that both measurement probabilities are non-zero
        if p_0 > float_tresh and p_1> float_tresh:
            
            #Normalize the new statevector arrays
            normalization = 1/(p_0**0.5)
            lower_half.data *= normalization
            
            #Create tensor factor corresponding to outcome 0
            tensor_factor_0 = TensorFactor(new_qubit_list, lower_half)
            
            
            #Perform the same operation for outcome 1
            normalization = 1/(p_1**0.5)
            upper_half.data *= normalization
            
            tensor_factor_1 = TensorFactor(list(new_qubit_list), upper_half)
        
        #This treats the cases that one of the outcomes has 0 probability
        elif p_0 < float_tresh:
            p_0 = 0
            p_1 = 1
            tensor_factor_0 = None
            tensor_factor_1 = TensorFactor(new_qubit_list, upper_half)
            
        else:
            p_0 = 1
            p_1 = 0
            tensor_factor_0 = TensorFactor(new_qubit_list, lower_half)
            tensor_factor_1 = None
        
        return p_0, tensor_factor_0, p_1, tensor_factor_1
    
    def unravel(self):
        
        sorted_qubit_list = list(self.qubits)
        sorted_qubit_list.sort()
        for i in range(len(self.qubits)):
            if sorted_qubit_list[i] == self.qubits[i]:
                continue
            self.swap(i, self.qubits.index(sorted_qubit_list[i]))



#This functions allow entangling more than two TensorFactors
#Since we are free to choose a contraction order, we pick one
#that synergizes with the parallelization features of the BiArrays
#ie. we entangle two factors, and put the entangled result at 
#the back of a queue. The processing of the entangling is the hopefully
#done once this factor is at the front of the queue.
def multi_entangle(input_factor_list):
    
    #If there is only one factor, there is nothing to be done
    if len(input_factor_list) == 1:
        return input_factor_list[0]
    
    else:
        #In the case of multiple factors, entangle the first two, and put the result at the back
        
        input_factor_list.sort(key = lambda x : len(x.qubits))
        input_factor_list.append(input_factor_list[0].entangle(input_factor_list[1]))
        
        #Remove entangling factors from the queue
        input_factor_list.pop(0)
        input_factor_list.pop(0)
        
        #Recurse on remaining factors
        return multi_entangle(input_factor_list)